import ast
import pathlib
from typing import List, Union

import pytest
from docs.sphinx_api_docs_source.include_exclude_definition import (
    IncludeExcludeDefinition,
)
from docs.sphinx_api_docs_source.public_api_report import (
    CodeParser,
    CodeReferenceFilter,
    Definition,
    DocsExampleParser,
    FileContents,
    PublicAPIChecker,
    PublicAPIReport,
    _get_import_names,
    get_shortest_dotted_path,
    parse_docs_contents_for_class_names,
)

pytestmark = pytest.mark.filesystem


@pytest.fixture
def sample_docs_example_python_file_string() -> str:
    return """from scripts.sample_with_definitions import ExampleClass, example_module_level_function

ec = ExampleClass()

ec.example_method()

a = ec.example_method_with_args(some_arg=1, other_arg=2)

b = ec.example_staticmethod()

c = ec.example_classmethod()

ec.example_public_staticmethod()

ec.example_public_classmethod()

d = ec.example_multiple_decorator_public_method()

ec._example_private_method()

ec.example_public_api_method()

e = example_module_level_function()

_example_private_module_level_function()

example_public_api_module_level_function()

epub = ExamplePublicAPIClass()

assert d

some_yaml_contents = \"\"\"
name: {datasource_name}
class_name: Something
\"\"\"

"""  # noqa: E501 # FIXME CoP


@pytest.fixture
def sample_with_definitions_python_file_string() -> str:
    return """
class ExampleClass:

    def __init__(self):
        pass

    def example_method(self):
        pass

    def example_method_with_args(self, some_arg, other_arg):
        pass

    @staticmethod
    def example_staticmethod():
        pass

    @classmethod
    def example_classmethod(cls):
        pass

    @staticmethod
    @public_api
    def example_public_staticmethod():
        pass

    @classmethod
    @public_api
    def example_public_classmethod(cls):
        pass

    @some_other_decorator
    @public_api
    @another_decorator
    def example_multiple_decorator_public_method(self):
        pass

    def _example_private_method(self):
        pass

    @public_api
    def example_public_api_method(self):
        pass

    def example_no_usages_in_sample_docs_example_python_file_string():
        pass


def example_module_level_function():
    pass

def _example_private_module_level_function():
    pass

@public_api
def example_public_api_module_level_function():
    pass

@public_api
class ExamplePublicAPIClass:
    pass
"""


@pytest.fixture
def sample_markdown_doc_with_yaml() -> str:
    return """# Title

Content.

More content.

Some yaml:

yaml_contents = \"\"\"
name: {datasource_name}
class_name: Something
\"\"\"

End of content.
"""


@pytest.fixture
def repo_root(tmp_path) -> pathlib.Path:
    return tmp_path


@pytest.fixture
def sample_docs_example_python_file_string_filepath(
    repo_root: pathlib.Path,
) -> pathlib.Path:
    path = (
        repo_root
        / pathlib.Path("tests/integration/docusaurus/sample_docs_example_python_file_string.py")
    ).relative_to(repo_root)
    path.touch()
    return path


@pytest.fixture
def sample_with_definitions_python_file_string_filepath(
    repo_root: pathlib.Path,
) -> pathlib.Path:
    path = (repo_root / pathlib.Path("sample_with_definitions_python_file_string.py")).relative_to(
        repo_root
    )
    path.touch()
    return path


@pytest.fixture
def sample_docs_example_file_contents(
    sample_docs_example_python_file_string: str,
    sample_docs_example_python_file_string_filepath: pathlib.Path,
) -> FileContents:
    return FileContents(
        filepath=sample_docs_example_python_file_string_filepath,
        contents=sample_docs_example_python_file_string,
    )


@pytest.fixture
def sample_with_definitions_file_contents(
    sample_with_definitions_python_file_string: str,
    sample_with_definitions_python_file_string_filepath: pathlib.Path,
) -> FileContents:
    return FileContents(
        filepath=sample_with_definitions_python_file_string_filepath,
        contents=sample_with_definitions_python_file_string,
    )


@pytest.fixture
def sample_markdown_doc_with_yaml_file_contents(
    sample_markdown_doc_with_yaml: str,
) -> FileContents:
    return FileContents(
        filepath=pathlib.Path("some/random/filepath/markdown.md"),
        contents=sample_markdown_doc_with_yaml,
    )


@pytest.fixture
def docs_example_parser(
    sample_docs_example_file_contents: FileContents,
) -> DocsExampleParser:
    docs_example_parser = DocsExampleParser(file_contents={sample_docs_example_file_contents})
    return docs_example_parser


@pytest.fixture
def empty_docs_example_parser(
    sample_docs_example_file_contents: FileContents,
) -> DocsExampleParser:
    docs_example_parser = DocsExampleParser(file_contents=set())
    return docs_example_parser


class TestDocExampleParser:
    def test_retrieve_all_usages_in_files(self, docs_example_parser: DocsExampleParser):
        usages = docs_example_parser.get_names_from_usage_in_docs_examples()
        assert usages == {
            "ExampleClass",
            "ExamplePublicAPIClass",
            "example_classmethod",
            "example_method",
            "example_method_with_args",
            "example_module_level_function",
            "example_multiple_decorator_public_method",
            "example_public_api_method",
            "example_public_api_module_level_function",
            "example_public_classmethod",
            "example_public_staticmethod",
            "example_staticmethod",
            "Something",
        }


@pytest.fixture
def code_parser(sample_with_definitions_file_contents: FileContents) -> CodeParser:
    code_parser = CodeParser(file_contents={sample_with_definitions_file_contents})
    return code_parser


class TestCodeParser:
    def test_get_all_class_method_and_function_names(self, code_parser: CodeParser):
        names = code_parser.get_all_class_method_and_function_names()
        assert names == {
            "ExampleClass",
            "ExamplePublicAPIClass",
            "__init__",
            "_example_private_method",
            "_example_private_module_level_function",
            "example_classmethod",
            "example_method",
            "example_method_with_args",
            "example_module_level_function",
            "example_multiple_decorator_public_method",
            "example_no_usages_in_sample_docs_example_python_file_string",
            "example_public_api_method",
            "example_public_api_module_level_function",
            "example_public_classmethod",
            "example_public_staticmethod",
            "example_staticmethod",
        }

    def test_get_all_class_method_and_function_definitions(self, code_parser: CodeParser):
        definitions = code_parser.get_all_class_method_and_function_definitions()

        assert len(definitions) == 16
        assert {d.name for d in definitions} == {
            "ExampleClass",
            "ExamplePublicAPIClass",
            "__init__",
            "_example_private_method",
            "_example_private_module_level_function",
            "example_classmethod",
            "example_method",
            "example_method_with_args",
            "example_module_level_function",
            "example_multiple_decorator_public_method",
            "example_no_usages_in_sample_docs_example_python_file_string",
            "example_public_api_method",
            "example_public_api_module_level_function",
            "example_public_classmethod",
            "example_public_staticmethod",
            "example_staticmethod",
        }
        assert {d.filepath for d in definitions} == {
            pathlib.Path("sample_with_definitions_python_file_string.py")
        }


def test_parse_docs_contents_for_class_names(
    sample_markdown_doc_with_yaml_file_contents: FileContents,
):
    assert parse_docs_contents_for_class_names(
        file_contents={sample_markdown_doc_with_yaml_file_contents}
    ) == {"Something"}


def test_get_shortest_dotted_path(monkeypatch):
    """Test path traversal using an example file.

    The example is just any class that is imported in a __init__.py file in a
    parent folder. This test can be modified to use any such import e.g. if the
    example file changes, or a fixture created so the example isn't coming
    from the repo.
    """
    repo_root = pathlib.Path(__file__).parent.parent.parent
    monkeypatch.chdir(repo_root)
    filepath = pathlib.Path("great_expectations/core/expectation_suite.py")
    definition = Definition(
        name="ExpectationSuite", filepath=filepath, ast_definition=ast.ClassDef()
    )

    # This is the actual path
    assert (
        get_shortest_dotted_path(definition=definition, repo_root_path=repo_root)
        != "great_expectations.core.expectation_suite.ExpectationSuite"
    )
    # This is the shortest path
    assert (
        get_shortest_dotted_path(definition=definition, repo_root_path=repo_root)
        == "great_expectations.ExpectationSuite"
    )


@pytest.fixture
def various_imports() -> str:
    return """import some_module
import SomeClass
from some_module import ClassFromSomeModule
from some_other_module import ClassFromSomeOtherModule
import AliasClass as ac
import alias_module as am
from alias_module import AliasClassFromAliasModule as acfam
from a.b.c import some_method as sm
"""


def test__get_import_names(various_imports: str):
    """Make sure the actual class and module names are returned."""
    tree = ast.parse(various_imports)
    import_names = []

    for node in ast.walk(tree):
        if isinstance(node, (ast.Import, ast.ImportFrom)):
            import_names.extend(_get_import_names(node))  # type: ignore[arg-type] # FIXME CoP

    assert import_names == [
        "some_module",
        "SomeClass",
        "ClassFromSomeModule",
        "ClassFromSomeOtherModule",
        "AliasClass",
        "alias_module",
        "AliasClassFromAliasModule",
        "some_method",
    ]


@pytest.fixture
def public_api_checker(code_parser: CodeParser) -> PublicAPIChecker:
    return PublicAPIChecker(code_parser=code_parser)


class TestPublicAPIChecker:
    def test_get_all_public_api_definitions(self, public_api_checker: PublicAPIChecker):
        observed = public_api_checker.get_all_public_api_definitions()
        assert len(observed) == 6
        assert {d.name for d in observed} == {
            "ExamplePublicAPIClass",
            "example_multiple_decorator_public_method",
            "example_public_api_method",
            "example_public_api_module_level_function",
            "example_public_classmethod",
            "example_public_staticmethod",
        }
        assert {d.filepath for d in observed} == {
            pathlib.Path("sample_with_definitions_python_file_string.py")
        }

    def _class_and_function_definitions(
        self, tree: ast.AST
    ) -> List[Union[ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef]]:
        """Helper function to find class and function definitions from ast tree for tests."""
        definitions = []
        for node in ast.walk(tree):
            if isinstance(node, (ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef)):
                definitions.append(node)

        return definitions

    def test_is_definition_marked_public_api_yes(
        self, public_api_checker: PublicAPIChecker, tmp_path: pathlib.Path
    ):
        file_string = """
@public_api
def example_public_api_module_level_function():
    pass

@public_api
class ExamplePublicAPIClass:
    @public_api
    def example_public_api_method():
        pass

    @staticmethod
    @public_api
    def example_public_api_staticmethod():
        pass

    @classmethod
    @public_api
    def example_public_api_classmethod(cls):
        pass

    @some_other_decorator
    @public_api
    @another_decorator
    def example_multiple_decorator_public_api_method(self):
        pass

"""
        test_path = tmp_path / "test_path.py"
        ast_definitions = self._class_and_function_definitions(tree=ast.parse(file_string))
        definitions = [
            Definition(
                name="test_name",
                filepath=pathlib.Path(test_path),
                ast_definition=ast_definition,
            )
            for ast_definition in ast_definitions
        ]
        assert all(
            public_api_checker.is_definition_marked_public_api(definition)
            for definition in definitions
        )

    def test_is_definition_marked_public_api_no(
        self, public_api_checker: PublicAPIChecker, tmp_path: pathlib.Path
    ):
        file_string = """
def example_module_level_function():
    pass

class ExampleClass:
    def example_method():
        pass

    @staticmethod
    def example_public_staticmethod():
        pass

    @classmethod
    def example_public_classmethod(cls):
        pass

    @some_other_decorator
    @another_decorator
    def example_multiple_decorator_public_method(self):
        pass

"""
        test_path = tmp_path / "test_path.py"
        ast_definitions = self._class_and_function_definitions(tree=ast.parse(file_string))
        definitions = [
            Definition(
                name="test_name",
                filepath=pathlib.Path(test_path),
                ast_definition=ast_definition,
            )
            for ast_definition in ast_definitions
        ]
        assert not all(
            public_api_checker.is_definition_marked_public_api(definition)
            for definition in definitions
        )


@pytest.fixture
def code_reference_filter(
    repo_root: pathlib.Path,
    docs_example_parser: DocsExampleParser,
    code_parser: CodeParser,
    public_api_checker: PublicAPIChecker,
) -> CodeReferenceFilter:
    return CodeReferenceFilter(
        repo_root=repo_root,
        docs_example_parser=docs_example_parser,
        code_parser=code_parser,
        public_api_checker=public_api_checker,
    )


@pytest.fixture
def code_reference_filter_with_non_default_include_exclude(
    repo_root: pathlib.Path,
    docs_example_parser: DocsExampleParser,
    code_parser: CodeParser,
    public_api_checker: PublicAPIChecker,
    sample_docs_example_python_file_string_filepath: pathlib.Path,
) -> CodeReferenceFilter:
    return CodeReferenceFilter(
        repo_root=repo_root,
        docs_example_parser=docs_example_parser,
        code_parser=code_parser,
        public_api_checker=public_api_checker,
        includes=[
            IncludeExcludeDefinition(
                reason="test",
                name="test_name",
                filepath=sample_docs_example_python_file_string_filepath,
            )
        ],
        excludes=[
            IncludeExcludeDefinition(
                reason="test",
                name="test_name",
                filepath=sample_docs_example_python_file_string_filepath,
            )
        ],
    )


@pytest.fixture
def code_reference_filter_with_no_include_exclude(
    repo_root: pathlib.Path,
    docs_example_parser: DocsExampleParser,
    code_parser: CodeParser,
    public_api_checker: PublicAPIChecker,
) -> CodeReferenceFilter:
    return CodeReferenceFilter(
        repo_root=repo_root,
        docs_example_parser=docs_example_parser,
        code_parser=code_parser,
        public_api_checker=public_api_checker,
        includes=[],
        excludes=[],
    )


@pytest.fixture
def code_reference_filter_with_exclude_by_file(
    repo_root: pathlib.Path,
    docs_example_parser: DocsExampleParser,
    code_parser: CodeParser,
    public_api_checker: PublicAPIChecker,
    sample_with_definitions_python_file_string_filepath: pathlib.Path,
) -> CodeReferenceFilter:
    return CodeReferenceFilter(
        repo_root=repo_root,
        docs_example_parser=docs_example_parser,
        code_parser=code_parser,
        public_api_checker=public_api_checker,
        includes=[],
        excludes=[
            IncludeExcludeDefinition(
                reason="test",
                filepath=sample_with_definitions_python_file_string_filepath,
            )
        ],
    )


@pytest.fixture
def code_reference_filter_with_references_from_docs_content(
    repo_root: pathlib.Path,
    empty_docs_example_parser: DocsExampleParser,
    code_parser: CodeParser,
    public_api_checker: PublicAPIChecker,
) -> CodeReferenceFilter:
    return CodeReferenceFilter(
        repo_root=repo_root,
        docs_example_parser=empty_docs_example_parser,
        code_parser=code_parser,
        public_api_checker=public_api_checker,
        references_from_docs_content={"ExampleClass", "ExamplePublicAPIClass"},
    )


@pytest.fixture
def code_reference_filter_with_exclude_by_file_and_name(
    repo_root: pathlib.Path,
    docs_example_parser: DocsExampleParser,
    code_parser: CodeParser,
    public_api_checker: PublicAPIChecker,
    sample_with_definitions_python_file_string_filepath: pathlib.Path,
) -> CodeReferenceFilter:
    return CodeReferenceFilter(
        repo_root=repo_root,
        docs_example_parser=docs_example_parser,
        code_parser=code_parser,
        public_api_checker=public_api_checker,
        includes=[],
        excludes=[
            IncludeExcludeDefinition(
                reason="test",
                name="example_method",
                filepath=sample_with_definitions_python_file_string_filepath,
            ),
            IncludeExcludeDefinition(
                reason="test",
                name="example_module_level_function",
                filepath=sample_with_definitions_python_file_string_filepath,
            ),
        ],
    )


@pytest.fixture
def code_reference_filter_with_include_by_file_and_name_already_included(
    repo_root: pathlib.Path,
    docs_example_parser: DocsExampleParser,
    code_parser: CodeParser,
    public_api_checker: PublicAPIChecker,
    sample_with_definitions_python_file_string_filepath: pathlib.Path,
) -> CodeReferenceFilter:
    return CodeReferenceFilter(
        repo_root=repo_root,
        docs_example_parser=docs_example_parser,
        code_parser=code_parser,
        public_api_checker=public_api_checker,
        includes=[
            IncludeExcludeDefinition(
                reason="test",
                name="example_method",
                filepath=sample_with_definitions_python_file_string_filepath,
            ),
            IncludeExcludeDefinition(
                reason="test",
                name="example_module_level_function",
                filepath=sample_with_definitions_python_file_string_filepath,
            ),
        ],
        excludes=[],
    )


@pytest.fixture
def code_reference_filter_with_include_by_file_and_name_already_excluded(
    repo_root: pathlib.Path,
    docs_example_parser: DocsExampleParser,
    code_parser: CodeParser,
    public_api_checker: PublicAPIChecker,
    sample_with_definitions_python_file_string_filepath: pathlib.Path,
) -> CodeReferenceFilter:
    return CodeReferenceFilter(
        repo_root=repo_root,
        docs_example_parser=docs_example_parser,
        code_parser=code_parser,
        public_api_checker=public_api_checker,
        includes=[
            IncludeExcludeDefinition(
                reason="test",
                name="example_method",
                filepath=sample_with_definitions_python_file_string_filepath,
            ),
            IncludeExcludeDefinition(
                reason="test",
                name="example_module_level_function",
                filepath=sample_with_definitions_python_file_string_filepath,
            ),
        ],
        excludes=[
            IncludeExcludeDefinition(
                reason="test",
                filepath=sample_with_definitions_python_file_string_filepath,
            )
        ],
    )


@pytest.fixture
def code_reference_filter_with_include_by_file_and_name_not_used_in_docs_example_exclude_file(
    repo_root: pathlib.Path,
    docs_example_parser: DocsExampleParser,
    code_parser: CodeParser,
    public_api_checker: PublicAPIChecker,
    sample_with_definitions_python_file_string_filepath: pathlib.Path,
) -> CodeReferenceFilter:
    return CodeReferenceFilter(
        repo_root=repo_root,
        docs_example_parser=docs_example_parser,
        code_parser=code_parser,
        public_api_checker=public_api_checker,
        includes=[
            IncludeExcludeDefinition(
                reason="test",
                name="example_no_usages_in_sample_docs_example_python_file_string",
                filepath=sample_with_definitions_python_file_string_filepath,
            ),
        ],
        excludes=[
            IncludeExcludeDefinition(
                reason="test",
                filepath=sample_with_definitions_python_file_string_filepath,
            )
        ],
    )


class TestCodeReferenceFilter:
    def test_instantiate(self, code_reference_filter: CodeReferenceFilter):
        assert code_reference_filter.excludes
        assert code_reference_filter.includes

    def test_instantiate_with_non_default_include_exclude(
        self,
        code_reference_filter_with_non_default_include_exclude: CodeReferenceFilter,
    ):
        code_reference_filter = code_reference_filter_with_non_default_include_exclude
        assert code_reference_filter.excludes
        assert code_reference_filter.includes
        assert len(code_reference_filter.excludes) == 1
        assert len(code_reference_filter.includes) == 1

    def test_filter_definitions_no_include_exclude(
        self, code_reference_filter_with_no_include_exclude: CodeReferenceFilter
    ):
        observed = code_reference_filter_with_no_include_exclude.filter_definitions()
        assert len(observed) == 6
        assert {d.name for d in observed} == {
            "ExampleClass",
            # "__init__",  # Filtered private methods
            # "_example_private_method",  # Filtered private methods
            # "_example_private_module_level_function",  # Filtered private methods
            "example_classmethod",
            "example_method",
            "example_method_with_args",
            "example_module_level_function",
            "example_staticmethod",
        }
        assert {d.filepath for d in observed} == {
            pathlib.Path("sample_with_definitions_python_file_string.py")
        }

    def test_filter_definitions_with_references_from_docs_content(
        self,
        code_reference_filter_with_references_from_docs_content: CodeReferenceFilter,
    ):
        observed = code_reference_filter_with_references_from_docs_content.filter_definitions()
        assert len(observed) == 1
        assert {d.name for d in observed} == {"ExampleClass"}
        assert {d.filepath for d in observed} == {
            pathlib.Path("sample_with_definitions_python_file_string.py")
        }

    def test_filter_definitions_exclude_by_file(
        self, code_reference_filter_with_exclude_by_file: CodeReferenceFilter
    ):
        observed = code_reference_filter_with_exclude_by_file.filter_definitions()
        assert len(observed) == 0
        assert {d.name for d in observed} == set()
        assert {d.filepath for d in observed} == set()

    def test_filter_definitions_exclude_by_file_and_name(
        self, code_reference_filter_with_exclude_by_file_and_name: CodeReferenceFilter
    ):
        observed = code_reference_filter_with_exclude_by_file_and_name.filter_definitions()
        assert len(observed) == 4
        assert {d.name for d in observed} == {
            "ExampleClass",
            "example_classmethod",
            "example_method_with_args",
            "example_staticmethod",
        }
        assert {d.filepath for d in observed} == {
            pathlib.Path("sample_with_definitions_python_file_string.py")
        }

    def test_filter_definitions_include_by_file_and_name_already_included(
        self,
        code_reference_filter_with_include_by_file_and_name_already_included: CodeReferenceFilter,
    ):
        """What does this test and why?

        That include directives that try to include already included definitions
        will not include multiple copies of the same definitions (when not
        accounting for different but equivalent ast definition object instances).
        """
        observed = code_reference_filter_with_include_by_file_and_name_already_included.filter_definitions()  # noqa: E501 # FIXME CoP
        # There are two extra (8 vs 6) here due to the ast_definition classes
        #  pointing to different but equivalent objects.
        assert len(observed) == 8
        assert {d.name for d in observed} == {
            "ExampleClass",
            "example_classmethod",
            "example_method",
            "example_method_with_args",
            "example_module_level_function",
            "example_staticmethod",
        }
        assert {d.filepath for d in observed} == {
            pathlib.Path("sample_with_definitions_python_file_string.py")
        }

    def test_filter_definitions_include_by_file_and_name_already_excluded(
        self,
        code_reference_filter_with_include_by_file_and_name_already_excluded: CodeReferenceFilter,
    ):
        """What does this test and why?

        Include overrides exclude.
        """
        observed = code_reference_filter_with_include_by_file_and_name_already_excluded.filter_definitions()  # noqa: E501 # FIXME CoP
        # There are two extra (4 vs 2) here due to the ast_definition classes
        #  pointing to different but equivalent objects.
        assert len(observed) == 4
        assert {d.name for d in observed} == {
            "example_method",
            "example_module_level_function",
        }
        assert {d.filepath for d in observed} == {
            pathlib.Path("sample_with_definitions_python_file_string.py")
        }

    def test_filter_definitions_include_by_file_and_name_already_excluded_not_used_in_docs_example(
        self,
        code_reference_filter_with_include_by_file_and_name_not_used_in_docs_example_exclude_file: CodeReferenceFilter,  # noqa: E501 # FIXME CoP
    ):
        """What does this test and why?

        Include overrides exclude. Method that was not included in docs examples
        is still included if manually added.
        """
        observed = code_reference_filter_with_include_by_file_and_name_not_used_in_docs_example_exclude_file.filter_definitions()  # noqa: E501 # FIXME CoP
        assert len(observed) == 1
        assert {d.name for d in observed} == {
            "example_no_usages_in_sample_docs_example_python_file_string",
        }
        assert {d.filepath for d in observed} == {
            pathlib.Path("sample_with_definitions_python_file_string.py")
        }


@pytest.fixture
def public_api_report(
    code_reference_filter_with_no_include_exclude: CodeReferenceFilter,
    repo_root: pathlib.Path,
) -> PublicAPIReport:
    return PublicAPIReport(
        definitions=code_reference_filter_with_no_include_exclude.filter_definitions(),
        repo_root=repo_root,
    )


@pytest.fixture
def public_api_report_filter_out_file(
    code_reference_filter_with_exclude_by_file: CodeReferenceFilter,
    repo_root: pathlib.Path,
) -> PublicAPIReport:
    return PublicAPIReport(
        definitions=code_reference_filter_with_exclude_by_file.filter_definitions(),
        repo_root=repo_root,
    )


class TestPublicAPIReport:
    def test_generate_printable_definitions(self, public_api_report: PublicAPIReport):
        expected: List[str] = [
            "File: sample_with_definitions_python_file_string.py Name: ExampleClass",
            "File: sample_with_definitions_python_file_string.py Name: example_classmethod",
            "File: sample_with_definitions_python_file_string.py Name: example_method",
            "File: sample_with_definitions_python_file_string.py Name: example_method_with_args",
            "File: sample_with_definitions_python_file_string.py Name: "
            "example_module_level_function",
            "File: sample_with_definitions_python_file_string.py Name: example_staticmethod",
        ]
        observed = [str(p) for p in public_api_report.generate_printable_definitions()]
        assert observed == expected

    def test_generate_printable_definitions_exclude_by_file(
        self, public_api_report_filter_out_file: PublicAPIReport
    ):
        expected: List[str] = []
        observed = public_api_report_filter_out_file.generate_printable_definitions()
        assert observed == expected


class TestIncludeExcludeDefinition:
    def test_instantiate_name_and_filepath(self, tmp_path: pathlib.Path):
        path = tmp_path / "test_path.py"
        path.touch()
        IncludeExcludeDefinition(
            reason="reason",
            name="name",
            filepath=path,
        )

    def test_instantiate_filepath_only(self, tmp_path: pathlib.Path):
        path = tmp_path / "test_path.py"
        path.touch()
        IncludeExcludeDefinition(reason="reason", filepath=path)

    def test_instantiate_name_only(self):
        with pytest.raises(ValueError) as exc:
            IncludeExcludeDefinition(reason="reason", name="name")

        assert "You must provide a filepath if also providing a name" in exc.value.args[0]

    def test_instantiate_reason_only(self):
        with pytest.raises(ValueError) as exc:
            IncludeExcludeDefinition(reason="reason")

        assert "You must provide at least a filepath or filepath and name" in exc.value.args[0]



import great_expectations as gx

context = gx.get_context()

datasource_name = "my_gcs_datasource"
bucket_or_name = "test_docs_data"

datasource = context.data_sources.add_pandas_gcs(
    name=datasource_name, bucket_or_name=bucket_or_name, gcs_options={}
)

assert datasource_name in context.data_sources.all()

asset_name = "my_taxi_data_asset"
gcs_prefix = "data/taxi_yellow_tripdata_samples/"
data_asset = datasource.add_csv_asset(
    name=asset_name,
    gcs_prefix=gcs_prefix,
)

batch_definition = data_asset.add_batch_definition_path(
    "path", path="yellow_tripdata_sample_2019-02.csv"
)

my_batch_request = batch_definition.build_batch_request()
batches = data_asset.get_batch(my_batch_request)
assert batches.metadata == {
    "path": "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-02.csv"
}



from __future__ import annotations

import logging
import textwrap
from typing import TYPE_CHECKING

from great_expectations.compatibility.typing_extensions import override

logger = logging.getLogger(__name__)

from great_expectations.render.renderer.renderer import Renderer

if TYPE_CHECKING:
    from great_expectations.checkpoint.checkpoint import CheckpointResult
    from great_expectations.core.expectation_validation_result import (
        ExpectationSuiteValidationResult,
    )


class EmailRenderer(Renderer):
    @override
    def render(self, checkpoint_result: CheckpointResult) -> tuple[str, str]:
        checkpoint_name = checkpoint_result.checkpoint_config.name
        status = checkpoint_result.success
        title = f"{checkpoint_name}: {status}"

        text_blocks: list[str] = []
        for result in checkpoint_result.run_results.values():
            html = self._render_validation_result(result=result)
            text_blocks.append(html)

        return title, self._concatenate_blocks(text_blocks=text_blocks)

    def _render_validation_result(self, result: ExpectationSuiteValidationResult) -> str:
        suite_name = result.suite_name
        asset_name = result.asset_name or "__no_asset_name__"
        n_checks_succeeded = result.statistics["successful_expectations"]
        n_checks = result.statistics["evaluated_expectations"]
        run_id = result.meta.get("run_id", "__no_run_id__")
        batch_id = result.batch_id
        check_details_text = f"<strong>{n_checks_succeeded}</strong> of <strong>{n_checks}</strong> expectations were met"  # noqa: E501 # FIXME CoP
        status = "Success 🎉" if result.success else "Failed ❌"

        title = f"<h3><u>{suite_name}</u></h3>"
        html = textwrap.dedent(
            f"""\
            <p><strong>{title}</strong></p>
            <p><strong>Batch Validation Status</strong>: {status}</p>
            <p><strong>Expectation Suite Name</strong>: {suite_name}</p>
            <p><strong>Data Asset Name</strong>: {asset_name}</p>
            <p><strong>Run ID</strong>: {run_id}</p>
            <p><strong>Batch ID</strong>: {batch_id}</p>
            <p><strong>Summary</strong>: {check_details_text}</p>"""
        )

        return html

    def _concatenate_blocks(self, text_blocks: list[str]) -> str:
        return "<br>".join(text_blocks)

    def _get_report_element(self, docs_link):
        report_element = None
        if docs_link:
            try:
                if "file://" in docs_link:
                    # handle special case since the email does not render these links
                    report_element = str(
                        f'<p><strong>DataDocs</strong> can be found here: <a href="{docs_link}">{docs_link}</a>.</br>'  # noqa: E501 # FIXME CoP
                        "(Please copy and paste link into a browser to view)</p>",
                    )
                else:
                    report_element = f'<p><strong>DataDocs</strong> can be found here: <a href="{docs_link}">{docs_link}</a>.</p>'  # noqa: E501 # FIXME CoP
            except Exception as e:
                logger.warning(
                    f"""EmailRenderer had a problem with generating the docs link.
                    link used to generate the docs link is: {docs_link} and is of type: {type(docs_link)}.
                    Error: {e}"""  # noqa: E501 # FIXME CoP
                )
                return
        else:
            logger.warning("No docs link found. Skipping data docs link in the email message.")
        return report_element



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from great_expectations.experimental.metric_repository.data_store import DataStore
    from great_expectations.experimental.metric_repository.metrics import MetricRun


class MetricRepository:
    """A repository for storing and retrieving MetricRuns.

    Args:
        data_store: The DataStore to use for storing and retrieving MetricRuns.
    """

    def __init__(self, data_store: DataStore):
        self._data_store = data_store

    def add_metric_run(self, metric_run: MetricRun) -> uuid.UUID:
        return self._data_store.add(value=metric_run)



import os

import great_expectations as gx

context = gx.get_context()

datasource_name = "ABS datasource"

CREDENTIAL = os.getenv("AZURE_CREDENTIAL", "")
ACCOUNT_URL = "superconductivetesting.blob.core.windows.net"
CONTAINER = "superconductive-public"
NAME_STARTS_WITH = "data/taxi_yellow_tripdata_samples/"


datasource = context.data_sources.add_pandas_abs(
    name=datasource_name,
    azure_options={
        "account_url": ACCOUNT_URL,
        "credential": CREDENTIAL,
    },
)

asset = datasource.add_csv_asset(
    name="taxi_yellow_tripdata_samples",
    abs_container=CONTAINER,
    abs_name_starts_with=NAME_STARTS_WITH,
)

batch_definition = asset.add_batch_definition_path(
    "abs batch definition",
    path="yellow_tripdata_sample_2019-02.csv",
)

batch_request = batch_definition.build_batch_request()
batch = asset.get_batch(batch_request)

assert batch.metadata == {
    "path": "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-02.csv"
}



from typing import Dict, List

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core import ExpectationSuite
from great_expectations.datasource.fluent import BatchRequest as FluentBatchRequest


@pytest.fixture
def update_data_docs_action():
    return {
        "name": "update_data_docs",
        "action": {"class_name": "UpdateDataDocsAction"},
    }


@pytest.fixture
def store_validation_result_action():
    return {
        "name": "store_validation_result",
        "action": {"class_name": "StoreValidationResultAction"},
    }


@pytest.fixture
def webhook() -> str:
    return "https://hooks.slack.com/foo/bar"


@pytest.fixture
def slack_notification_action(webhook):
    return {
        "name": "send_slack_notification",
        "action": {
            "class_name": "SlackNotificationAction",
            "slack_webhook": webhook,
            "notify_on": "all",
            "notify_with": None,
            "renderer": {
                "module_name": "great_expectations.render.renderer.slack_renderer",
                "class_name": "SlackRenderer",
            },
        },
    }


@pytest.fixture
def common_action_list(
    store_validation_result_action: dict,
    update_data_docs_action: dict,
) -> List[dict]:
    return [
        store_validation_result_action,
        update_data_docs_action,
    ]


@pytest.fixture
def batch_request_as_dict() -> Dict[str, str]:
    return {
        "datasource_name": "my_pandas_filesystem_datasource",
        "data_asset_name": "users",
    }


@pytest.fixture
def fluent_batch_request(batch_request_as_dict: Dict[str, str]) -> FluentBatchRequest:
    return FluentBatchRequest(
        datasource_name=batch_request_as_dict["datasource_name"],
        data_asset_name=batch_request_as_dict["data_asset_name"],
    )


@pytest.fixture
def titanic_pandas_data_context_stats_enabled_and_expectation_suite_with_one_expectation(
    titanic_pandas_data_context_with_v013_datasource_with_checkpoints_v1_with_empty_store_stats_enabled,
):
    context = titanic_pandas_data_context_with_v013_datasource_with_checkpoints_v1_with_empty_store_stats_enabled  # noqa: E501 # FIXME CoP
    # create expectation suite
    suite = context.suites.add(ExpectationSuite("my_expectation_suite"))
    expectation = gxe.ExpectColumnValuesToBeBetween(
        column="col1",
        min_value=1,
        max_value=2,
    )
    suite.add_expectation(expectation=expectation)
    # noinspection PyProtectedMember
    context._save_project_config()
    return context


@pytest.fixture
def titanic_data_context_with_fluent_pandas_datasources_stats_enabled_and_expectation_suite_with_one_expectation(  # noqa: E501 # FIXME CoP
    titanic_data_context_with_fluent_pandas_datasources_with_checkpoints_v1_with_empty_store_stats_enabled,
):
    context = titanic_data_context_with_fluent_pandas_datasources_with_checkpoints_v1_with_empty_store_stats_enabled  # noqa: E501 # FIXME CoP

    datasource_name = "my_pandas_filesystem_datasource"
    datasource = context.data_sources.get(datasource_name=datasource_name)

    batching_regex = r"^Titanic_1911\.csv"
    glob_directive = "*.csv"
    datasource.add_csv_asset(
        name="Titanic_1911",
        batching_regex=batching_regex,
        glob_directive=glob_directive,
    )

    datasource_name = "my_pandas_dataframes_datasource"
    datasource = context.data_sources.get(datasource_name=datasource_name)

    test_df: pd.DataFrame = pd.DataFrame(data={"col1": [1, 2], "col2": [3, 4]})
    asset = datasource.add_dataframe_asset(name="my_other_dataframe_asset")
    _ = asset.build_batch_request(options={"dataframe": test_df})

    # create expectation suite
    suite = context.suites.add(ExpectationSuite("my_expectation_suite"))
    expectation = gxe.ExpectColumnValuesToBeBetween(
        column="col1",
        min_value=1,
        max_value=2,
    )
    suite.add_expectation(expectation)
    # noinspection PyProtectedMember
    context._save_project_config()

    return context


@pytest.fixture
def titanic_data_context_with_fluent_pandas_and_spark_datasources_stats_enabled_and_expectation_suite_with_one_expectation(  # noqa: E501 # FIXME CoP
    titanic_data_context_with_fluent_pandas_and_spark_datasources_with_checkpoints_v1_with_empty_store_stats_enabled,
):
    context = titanic_data_context_with_fluent_pandas_and_spark_datasources_with_checkpoints_v1_with_empty_store_stats_enabled  # noqa: E501 # FIXME CoP

    datasource_name = "my_pandas_filesystem_datasource"
    datasource = context.data_sources.get(datasource_name=datasource_name)

    batching_regex = r"^Titanic_1911\.csv"
    glob_directive = "*.csv"
    datasource.add_csv_asset(
        name="Titanic_1911",
        batching_regex=batching_regex,
        glob_directive=glob_directive,
    )

    # create expectation suite
    suite = context.suites.add(ExpectationSuite("my_expectation_suite"))
    expectation = gxe.ExpectColumnValuesToBeBetween(
        column="col1",
        min_value=1,
        max_value=2,
    )
    suite.add_expectation(expectation)
    # noinspection PyProtectedMember
    context._save_project_config()

    return context



from types import ModuleType
from typing import List, Optional

import pytest

from great_expectations.expectations.metrics import ColumnValuesInSet

try:
    import sqlalchemy
    from sqlalchemy.dialects import mysql, postgresql, sqlite
except ImportError:
    sqlalchemy = None

try:
    import sqlalchemy_bigquery
except ImportError:
    sqlalchemy_bigquery = None


@pytest.mark.unit
@pytest.mark.skipif(
    sqlalchemy is None or sqlalchemy_bigquery is None,
    reason="sqlalchemy or sqlalchemy_bigquery is not installed",
)
@pytest.mark.parametrize("value_set", [[False, True], [False, True, None], [True], [False], [None]])
def test_sqlalchemy_impl_bigquery_bool(value_set: List[Optional[bool]]):
    column_name = "my_bool_col"
    column = sqlalchemy.column(column_name)
    kwargs = _make_sqlalchemy_kwargs(column_name, sqlalchemy_bigquery)
    predicate = ColumnValuesInSet._sqlalchemy_impl(column, value_set, **kwargs)
    # If a value in value_set is None we expect "column_name is null" otherwise we expect
    # "column_name = value"
    expected_predicates = [
        f"{column_name} {'is' if value is None else '='} {'null' if value is None else value}"
        for value in value_set
    ]
    assert str(predicate).lower() == " or ".join(expected_predicates).lower()


@pytest.mark.unit
@pytest.mark.skipif(sqlalchemy is None, reason="sqlalchemy is not installed")
@pytest.mark.parametrize("dialect", [mysql, postgresql, sqlite])
@pytest.mark.parametrize("value_set", [[False, True], [False, True, None], [True], [False], [None]])
def test_sqlalchemy_impl_not_bigquery_bool(dialect: ModuleType, value_set: List[Optional[bool]]):
    column_name = "my_bool_col"
    column = sqlalchemy.column(column_name)
    kwargs = _make_sqlalchemy_kwargs(column_name, dialect)
    predicate = ColumnValuesInSet._sqlalchemy_impl(column, value_set, **kwargs)
    expected_predicates = ", ".join(
        ["null" if value is None else str(value) for value in value_set]
    ).lower()
    # We expect the predicate to look similar to "column_name in (true, false, null)"
    assert (
        str(predicate.compile(compile_kwargs={"literal_binds": True})).lower()
        == f"{column_name} in ({expected_predicates})"
    )


def _make_sqlalchemy_kwargs(column_name, dialect):
    return {
        "_dialect": dialect,
        "_metrics": {
            "table.column_types": [
                {
                    "comment": None,
                    "default": None,
                    "max_length": None,
                    "name": column_name,
                    "nullable": True,
                    "precision": None,
                    "scale": None,
                    "type": sqlalchemy.Boolean(),
                },
            ],
            "table.columns": [column_name],
            "table.row_count": 10000,
        },
        "parse_strings_as_datetimes": False,
        # The following key/values to be passed into _sqlalchemy_impl, but they aren't
        # actually used
        "_table": sqlalchemy.Table("my_table", sqlalchemy.MetaData()),
        "_sqlalchemy_engine": "DummySqlalchemyEngine",
    }



"""Test cases and fixtures for sampler integration test configurations."""

from dataclasses import dataclass
from typing import List

import pandas as pd


class SamplerTaxiTestData:
    def __init__(self, test_df: pd.DataFrame, test_column_name: str):
        self._test_df = test_df
        self._test_column_name = test_column_name

    @property
    def test_df(self):
        return self._test_df

    @property
    def test_column_name(self):
        return self._test_column_name

    def first_n_rows(self, n: int) -> pd.DataFrame:
        """Return first n rows of the test df.

        Args:
            n: Number of rows to include.

        Returns:
            The first n rows of the loaded test dataframe.
        """
        return self.test_df.head(n=n)


@dataclass
class TaxiSamplingTestCase:
    sampling_method_name: str
    sampling_kwargs: dict
    num_expected_batch_definitions: int
    num_expected_rows_in_first_batch_definition: int


class TaxiSamplingTestCases:
    def __init__(self, taxi_test_data: SamplerTaxiTestData):
        self._taxi_test_data = taxi_test_data

    @property
    def taxi_test_data(self) -> SamplerTaxiTestData:
        return self._taxi_test_data

    @property
    def test_df(self) -> pd.DataFrame:
        return self._taxi_test_data.test_df

    @property
    def test_column_name(self) -> str:
        return self._taxi_test_data.test_column_name

    def test_cases(self) -> List[TaxiSamplingTestCase]:
        return [
            TaxiSamplingTestCase(
                sampling_method_name="sample_using_limit",
                sampling_kwargs={"n": 10},
                num_expected_batch_definitions=1,
                num_expected_rows_in_first_batch_definition=10,
            ),
        ]



from __future__ import annotations

import datetime
import logging
import re
import warnings
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable, Mapping, MutableMapping, Optional, TypeVar, Union
from urllib.parse import urlparse

import dateutil.parser
import numpy as np

from great_expectations import exceptions as gx_exceptions
from great_expectations.compatibility.sqlalchemy import SQLALCHEMY_NOT_IMPORTED, LegacyRow, Row

if TYPE_CHECKING:
    from great_expectations.compatibility import pyspark


logger = logging.getLogger(__name__)

try:
    from shapely.geometry import LineString, MultiPolygon, Point, Polygon
except ImportError:
    Point = None
    Polygon = None
    MultiPolygon = None
    LineString = None


if not LegacyRow:
    LegacyRow = SQLALCHEMY_NOT_IMPORTED

if not Row:  # type: ignore[truthy-function] # FIXME CoP
    Row = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc] # FIXME CoP

SCHEMAS = {
    "api_np": {
        "NegativeInfinity": -np.inf,
        "PositiveInfinity": np.inf,
    },
    "api_cast": {
        "NegativeInfinity": -float("inf"),
        "PositiveInfinity": float("inf"),
    },
    "mysql": {
        "NegativeInfinity": -1.79e308,
        "PositiveInfinity": 1.79e308,
    },
    "mssql": {
        "NegativeInfinity": -1.79e308,
        "PositiveInfinity": 1.79e308,
    },
}


_SUFFIX_TO_PD_KWARG = {"gz": "gzip", "zip": "zip", "bz2": "bz2", "xz": "xz"}

M = TypeVar("M", bound=MutableMapping)


def nested_update(
    d: M,
    u: Mapping,
    dedup: bool = False,
    concat_lists: bool = True,
) -> M:
    """
    Update d with items from u, recursively and joining elements. By default, list values are
    concatenated without de-duplication. If concat_lists is set to False, lists in u (new dict)
    will replace those in d (base dict).
    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            d[k] = nested_update(d.get(k, {}), v, dedup=dedup)
        elif isinstance(v, set) or (k in d and isinstance(d[k], set)):
            s1 = d.get(k, set())
            s2 = v or set()

            if concat_lists:
                d[k] = s1 | s2
            else:
                d[k] = s2
        elif isinstance(v, list) or (k in d and isinstance(d[k], list)):
            l1 = d.get(k, [])
            l2 = v or []
            if concat_lists:
                if dedup:
                    d[k] = list(set(l1 + l2))
                else:
                    d[k] = l1 + l2
            else:
                d[k] = l2
        else:
            d[k] = v
    return d


def in_jupyter_notebook():
    try:
        from IPython import get_ipython  # type: ignore[import-not-found] # FIXME CoP

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except (NameError, ImportError):
        return False  # Probably standard Python interpreter


def determine_progress_bar_method_by_environment() -> Callable:
    """
    As tqdm has specific methods for progress bar creation and iteration,
    we require a utility to determine which method to use.

    If in a Jupyter notebook, we want to use `tqdm.notebook.tqdm`. Otherwise,
    we default to the standard `tqdm.tqdm`. Please see the docs for more information: https://tqdm.github.io/

    Returns:
        The appropriate tqdm method for the environment in question.
    """
    from tqdm import tqdm
    from tqdm.notebook import tqdm as tqdm_notebook

    if in_jupyter_notebook():
        return tqdm_notebook
    return tqdm


def substitute_all_strftime_format_strings(
    data: Union[dict, list, str, Any], datetime_obj: Optional[datetime.datetime] = None
) -> Union[str, Any]:
    """
    This utility function will iterate over input data and for all strings, replace any strftime format
    elements using either the provided datetime_obj or the current datetime
    """  # noqa: E501 # FIXME CoP

    datetime_obj = datetime_obj or datetime.datetime.now()  # noqa: DTZ005 # FIXME CoP
    if isinstance(data, (dict, OrderedDict)):
        return {
            k: substitute_all_strftime_format_strings(v, datetime_obj=datetime_obj)
            for k, v in data.items()
        }
    elif isinstance(data, list):
        return [
            substitute_all_strftime_format_strings(el, datetime_obj=datetime_obj) for el in data
        ]
    elif isinstance(data, str):
        return datetime_obj.strftime(data)
    else:
        return data


def parse_string_to_datetime(
    datetime_string: str, datetime_format_string: Optional[str] = None
) -> datetime.datetime:
    if not isinstance(datetime_string, str):
        raise gx_exceptions.SorterError(  # noqa: TRY003 # FIXME CoP
            f"""Source "datetime_string" must have string type (actual type is "{type(datetime_string)!s}").
            """  # noqa: E501 # FIXME CoP
        )

    if not datetime_format_string:
        return dateutil.parser.parse(timestr=datetime_string)

    if datetime_format_string and not isinstance(datetime_format_string, str):
        raise gx_exceptions.SorterError(  # noqa: TRY003 # FIXME CoP
            f"""DateTime parsing formatter "datetime_format_string" must have string type (actual type is
"{type(datetime_format_string)!s}").
            """  # noqa: E501 # FIXME CoP
        )

    return datetime.datetime.strptime(  # noqa: DTZ007 # FIXME CoP
        datetime_string, datetime_format_string
    )


def datetime_to_int(dt: datetime.date) -> int:
    return int(dt.strftime("%Y%m%d%H%M%S"))


# noinspection SpellCheckingInspection
class AzureUrl:
    """
    Parses an Azure Blob Storage URL into its separate components.
    Formats:
        WASBS (for Spark): "wasbs://<CONTAINER>@<ACCOUNT_NAME>.blob.core.windows.net/<BLOB>"
        HTTP(S) (for Pandas) "<ACCOUNT_NAME>.blob.core.windows.net/<CONTAINER>/<BLOB>"

        Reference: WASBS -- Windows Azure Storage Blob (https://datacadamia.com/azure/wasb).
    """

    AZURE_BLOB_STORAGE_PROTOCOL_DETECTION_REGEX_PATTERN: str = (
        r"^[^@]+@.+\.blob\.core\.windows\.net\/.+$"
    )

    AZURE_BLOB_STORAGE_HTTPS_URL_REGEX_PATTERN: str = (
        r"^(https?:\/\/)?(.+?)\.blob\.core\.windows\.net/([^/]+)/(.+)$"
    )
    AZURE_BLOB_STORAGE_HTTPS_URL_TEMPLATE: str = (
        "{account_name}.blob.core.windows.net/{container}/{path}"
    )

    AZURE_BLOB_STORAGE_WASBS_URL_REGEX_PATTERN: str = (
        r"^(wasbs?:\/\/)?([^/]+)@(.+?)\.blob\.core\.windows\.net/(.+)$"
    )
    AZURE_BLOB_STORAGE_WASBS_URL_TEMPLATE: str = (
        "wasbs://{container}@{account_name}.blob.core.windows.net/{path}"
    )

    def __init__(self, url: str) -> None:
        search = re.search(AzureUrl.AZURE_BLOB_STORAGE_PROTOCOL_DETECTION_REGEX_PATTERN, url)
        if search is None:
            search = re.search(AzureUrl.AZURE_BLOB_STORAGE_HTTPS_URL_REGEX_PATTERN, url)
            assert search is not None, (
                "The provided URL does not adhere to the format specified by the "
                "Azure SDK (<ACCOUNT_NAME>.blob.core.windows.net/<CONTAINER>/<BLOB>)"
            )
            self._protocol = search.group(1)
            self._account_name = search.group(2)
            self._container = search.group(3)
            self._blob = search.group(4)
        else:
            search = re.search(AzureUrl.AZURE_BLOB_STORAGE_WASBS_URL_REGEX_PATTERN, url)
            assert search is not None, (
                "The provided URL does not adhere to the format specified by the Azure SDK (wasbs://<CONTAINER>@<ACCOUNT_NAME>.blob.core.windows.net/<BLOB>)"
            )
            self._protocol = search.group(1)
            self._container = search.group(2)
            self._account_name = search.group(3)
            self._blob = search.group(4)

    @property
    def protocol(self):
        return self._protocol

    @property
    def account_name(self):
        return self._account_name

    @property
    def account_url(self):
        return f"{self.account_name}.blob.core.windows.net"

    @property
    def container(self):
        return self._container

    @property
    def blob(self):
        return self._blob


class GCSUrl:
    """
    Parses a Google Cloud Storage URL into its separate components
    Format: gs://<BUCKET_OR_NAME>/<BLOB>
    """

    URL_REGEX_PATTERN: str = r"^gs://([^/]+)/(.+)$"

    OBJECT_URL_TEMPLATE: str = "gs://{bucket_or_name}/{path}"

    def __init__(self, url: str) -> None:
        search = re.search(GCSUrl.URL_REGEX_PATTERN, url)
        assert search is not None, (
            "The provided URL does not adhere to the format specified by the GCS SDK (gs://<BUCKET_OR_NAME>/<BLOB>)"
        )
        self._bucket = search.group(1)
        self._blob = search.group(2)

    @property
    def bucket(self):
        return self._bucket

    @property
    def blob(self):
        return self._blob


# S3Url class courtesy: https://stackoverflow.com/questions/42641315/s3-urls-get-bucket-name-and-path
class S3Url:
    OBJECT_URL_TEMPLATE: str = "s3a://{bucket}/{path}"

    """
    >>> s = S3Url("s3://bucket/hello/world")
    >>> s.bucket
    'bucket'
    >>> s.key
    'hello/world'
    >>> s.url
    's3://bucket/hello/world'

    >>> s = S3Url("s3://bucket/hello/world?qwe1=3#ddd")
    >>> s.bucket
    'bucket'
    >>> s.key
    'hello/world?qwe1=3#ddd'
    >>> s.url
    's3://bucket/hello/world?qwe1=3#ddd'

    >>> s = S3Url("s3://bucket/hello/world#foo?bar=2")
    >>> s.key
    'hello/world#foo?bar=2'
    >>> s.url
    's3://bucket/hello/world#foo?bar=2'
    """

    def __init__(self, url) -> None:
        self._parsed = urlparse(url, allow_fragments=False)

    @property
    def bucket(self):
        return self._parsed.netloc

    @property
    def key(self):
        if self._parsed.query:
            return f"{self._parsed.path.lstrip('/')}?{self._parsed.query}"
        else:
            return self._parsed.path.lstrip("/")

    @property
    def suffix(self) -> Optional[str]:
        """
        Attempts to get a file suffix from the S3 key.
        If can't find one returns `None`.
        """
        splits = self._parsed.path.rsplit(".", 1)
        _suffix = splits[-1]
        if len(_suffix) > 0 and len(splits) > 1:
            return str(_suffix)
        return None

    @property
    def url(self):
        return self._parsed.geturl()


class DBFSPath:
    """
    Methods for converting Databricks Filesystem (DBFS) paths
    """

    @staticmethod
    def convert_to_file_semantics_version(path: str) -> str:
        if re.search(r"^dbfs:", path):
            return path.replace("dbfs:", "/dbfs", 1)

        if re.search("^/dbfs", path):
            return path

        raise ValueError("Path should start with either /dbfs or dbfs:")  # noqa: TRY003 # FIXME CoP

    @staticmethod
    def convert_to_protocol_version(path: str) -> str:
        if re.search(r"^\/dbfs", path):
            candidate = path.replace("/dbfs", "dbfs:", 1)
            if candidate == "dbfs:":
                # Must add trailing slash
                return "dbfs:/"

            return candidate

        if re.search(r"^dbfs:", path):
            if path == "dbfs:":
                # Must add trailing slash
                return "dbfs:/"

            return path

        raise ValueError("Path should start with either /dbfs or dbfs:")  # noqa: TRY003 # FIXME CoP


def sniff_s3_compression(s3_url: S3Url) -> Union[str, None]:
    """Attempts to get read_csv compression from s3_url"""
    return _SUFFIX_TO_PD_KWARG.get(s3_url.suffix) if s3_url.suffix else None


def get_or_create_spark_application(
    spark_config: Optional[dict[str, str]] = None,
    force_reuse_spark_context: Optional[bool] = None,
) -> pyspark.SparkSession:
    from great_expectations.execution_engine import SparkDFExecutionEngine

    # deprecated-v1.0.0
    warnings.warn(
        "Utility method get_or_create_spark_application() is deprecated and will be removed in v1.0.0. "  # noqa: E501 # FIXME CoP
        "Please pass your spark_config to the relevant Spark Datasource, or create your Spark Session outside of GX.",  # noqa: E501 # FIXME CoP
        category=DeprecationWarning,
    )
    if force_reuse_spark_context is not None:
        # deprecated-v1.0.0
        warnings.warn(
            "force_reuse_spark_context is deprecated and will be removed in version 1.0. "
            "In environments that allow it, the existing Spark context will be reused, adding the "
            "spark_config options that have been passed. If the Spark context cannot be updated with "  # noqa: E501 # FIXME CoP
            "the spark_config, the context will be stopped and restarted with the new spark_config.",  # noqa: E501 # FIXME CoP
            category=DeprecationWarning,
        )
    return SparkDFExecutionEngine.get_or_create_spark_session(
        spark_config=spark_config  # type:ignore[arg-type]
    )


def get_or_create_spark_session(
    spark_config: Optional[dict[str, str]] = None,
) -> pyspark.SparkSession:
    """Obtains Spark session if it already exists; otherwise creates Spark session and returns it to caller.

    Args:
        spark_config: Dictionary containing Spark configuration (string-valued keys mapped to string-valued properties).

    Returns:
        SparkSession
    """  # noqa: E501 # FIXME CoP
    from great_expectations.execution_engine import SparkDFExecutionEngine

    # deprecated-v1.0.0
    warnings.warn(
        "Utility method get_or_create_spark_session() is deprecated and will be removed in v1.0.0. "
        "Please pass your spark_config to the relevant Spark Datasource, or create your Spark Session outside of GX.",  # noqa: E501 # FIXME CoP
        category=DeprecationWarning,
    )

    return SparkDFExecutionEngine.get_or_create_spark_session(
        spark_config=spark_config or {},  # type: ignore[arg-type] # FIXME CoP
    )


def get_sql_dialect_floating_point_infinity_value(schema: str, negative: bool = False) -> float:
    res: Optional[dict] = SCHEMAS.get(schema)
    if res is None:
        if negative:
            return -np.inf
        else:
            return np.inf
    else:  # noqa: PLR5501 # FIXME CoP
        if negative:
            return res["NegativeInfinity"]
        else:
            return res["PositiveInfinity"]



"""Test using actual sample data."""

from __future__ import annotations

from typing import List

import pandas as pd
import pytest
from pandas import Timestamp

from great_expectations.data_context import CloudDataContext
from great_expectations.datasource.fluent.batch_request import BatchRequest
from great_expectations.experimental.metric_repository.metric_list_metric_retriever import (
    MetricListMetricRetriever,
)
from great_expectations.experimental.metric_repository.metrics import (
    ColumnMetric,
    MetricTypes,
    TableMetric,
)


@pytest.fixture
def cloud_context_and_batch_request_with_simple_dataframe(
    empty_cloud_context_fluent: CloudDataContext,  # used as a fixture
):
    context = empty_cloud_context_fluent
    datasource = context.data_sources.add_pandas(name="my_pandas_datasource")

    d = {
        "numeric_with_nulls_1": [1, 2, None],
        "numeric_with_nulls_2": [3, 4, None],
        "string": ["a", "b", "c"],
        "string_with_nulls": ["a", "b", None],
        "boolean": [True, False, True],
        "datetime": [
            pd.to_datetime("2020-01-01"),
            pd.to_datetime("2020-01-02"),
            pd.to_datetime("2020-01-03"),
        ],
    }
    df = pd.DataFrame(data=d)

    name = "dataframe"
    data_asset = datasource.add_dataframe_asset(name=name)
    batch_request = data_asset.build_batch_request(options={"dataframe": df})
    return context, batch_request


@pytest.mark.cloud
def test_get_metrics_table_metrics_only(
    cloud_context_and_batch_request_with_simple_dataframe: tuple[CloudDataContext, BatchRequest],
):
    context, batch_request = cloud_context_and_batch_request_with_simple_dataframe
    table_metrics_list: List[MetricTypes] = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
    ]
    metric_retriever = MetricListMetricRetriever(context)
    metrics = metric_retriever.get_metrics(
        batch_request=batch_request, metric_list=table_metrics_list
    )
    validator = context.get_validator(batch_request=batch_request)
    batch_id = validator.active_batch.id

    expected_metrics = [
        TableMetric[int](
            batch_id=batch_id,
            metric_name="table.row_count",
            value=3,
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id=batch_id,
            metric_name="table.columns",
            value=[
                "numeric_with_nulls_1",
                "numeric_with_nulls_2",
                "string",
                "string_with_nulls",
                "boolean",
                "datetime",
            ],
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id=batch_id,
            metric_name="table.column_types",
            value=[
                {"name": "numeric_with_nulls_1", "type": "float64"},
                {"name": "numeric_with_nulls_2", "type": "float64"},
                {"name": "string", "type": "object"},
                {"name": "string_with_nulls", "type": "object"},
                {"name": "boolean", "type": "bool"},
                {"name": "datetime", "type": "datetime64[ns]"},
            ],
            exception=None,
        ),
    ]

    # Assert each metric so it is easier to see which one fails (instead of assert metrics == expected_metrics):  # noqa: E501 # FIXME CoP
    assert len(metrics) == len(expected_metrics)
    for metric in metrics:
        assert metric.dict() in [expected_metric.dict() for expected_metric in expected_metrics]


@pytest.mark.cloud
def test_get_metrics_full_cdm(
    cloud_context_and_batch_request_with_simple_dataframe: tuple[CloudDataContext, BatchRequest],
):
    context, batch_request = cloud_context_and_batch_request_with_simple_dataframe
    cdm_metrics_list: List[MetricTypes] = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
        MetricTypes.COLUMN_MIN,
        MetricTypes.COLUMN_MAX,
        MetricTypes.COLUMN_MEAN,
        MetricTypes.COLUMN_MEDIAN,
        MetricTypes.COLUMN_NULL_COUNT,
    ]
    metric_retriever = MetricListMetricRetriever(context)
    metrics = metric_retriever.get_metrics(
        batch_request=batch_request, metric_list=cdm_metrics_list
    )
    validator = context.get_validator(batch_request=batch_request)
    batch_id = validator.active_batch.id

    expected_metrics = [
        TableMetric[int](
            batch_id=batch_id,
            metric_name="table.row_count",
            value=3,
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id=batch_id,
            metric_name="table.columns",
            value=[
                "numeric_with_nulls_1",
                "numeric_with_nulls_2",
                "string",
                "string_with_nulls",
                "boolean",
                "datetime",
            ],
            exception=None,
        ),
        ColumnMetric[float](
            batch_id=batch_id,
            metric_name="column.min",
            column="numeric_with_nulls_1",
            value=1,
            exception=None,
        ),
        ColumnMetric[float](
            batch_id=batch_id,
            metric_name="column.min",
            column="numeric_with_nulls_2",
            value=3,
            exception=None,
        ),
        ColumnMetric[float](
            batch_id=batch_id,
            metric_name="column.max",
            column="numeric_with_nulls_1",
            value=2,
            exception=None,
        ),
        ColumnMetric[float](
            batch_id=batch_id,
            metric_name="column.max",
            column="numeric_with_nulls_2",
            value=4,
            exception=None,
        ),
        ColumnMetric[float](
            batch_id=batch_id,
            metric_name="column.mean",
            column="numeric_with_nulls_1",
            value=1.5,
            exception=None,
        ),
        ColumnMetric[float](
            batch_id=batch_id,
            metric_name="column.mean",
            column="numeric_with_nulls_2",
            value=3.5,
            exception=None,
        ),
        ColumnMetric[float](
            batch_id=batch_id,
            metric_name="column.median",
            column="numeric_with_nulls_1",
            value=1.5,
            exception=None,
        ),
        ColumnMetric[float](
            batch_id=batch_id,
            metric_name="column.median",
            column="numeric_with_nulls_2",
            value=3.5,
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id=batch_id,
            metric_name="table.column_types",
            value=[
                {"name": "numeric_with_nulls_1", "type": "float64"},
                {"name": "numeric_with_nulls_2", "type": "float64"},
                {"name": "string", "type": "object"},
                {"name": "string_with_nulls", "type": "object"},
                {"name": "boolean", "type": "bool"},
                {"name": "datetime", "type": "datetime64[ns]"},
            ],
            exception=None,
        ),
        ColumnMetric[int](
            batch_id=batch_id,
            metric_name="column_values.null.count",
            column="numeric_with_nulls_1",
            value=1,
            exception=None,
        ),
        ColumnMetric[int](
            batch_id=batch_id,
            metric_name="column_values.null.count",
            column="numeric_with_nulls_2",
            value=1,
            exception=None,
        ),
        ColumnMetric[int](
            batch_id=batch_id,
            metric_name="column_values.null.count",
            column="string",
            value=0,
            exception=None,
        ),
        ColumnMetric[int](
            batch_id=batch_id,
            metric_name="column_values.null.count",
            column="string_with_nulls",
            value=1,
            exception=None,
        ),
        ColumnMetric[int](
            batch_id=batch_id,
            metric_name="column_values.null.count",
            column="boolean",
            value=0,
            exception=None,
        ),
        ColumnMetric[int](
            batch_id=batch_id,
            metric_name="column_values.null.count",
            column="datetime",
            value=0,
            exception=None,
        ),
        ColumnMetric[str](
            batch_id=batch_id,
            metric_name="column.min",
            value=Timestamp("2020-01-01 00:00:00"),
            exception=None,
            column="datetime",
        ),
        ColumnMetric[str](
            batch_id=batch_id,
            metric_name="column.max",
            value=Timestamp("2020-01-03 00:00:00"),
            exception=None,
            column="datetime",
        ),
    ]

    assert len(metrics) == len(expected_metrics)
    for metric in metrics:
        assert metric.dict() in [expected_metric.dict() for expected_metric in expected_metrics]



from .query_column import QueryColumn
from .query_column_pair import QueryColumnPair
from .query_data_source_table.comparison_query_data_source_table import (
    ComparisonQueryDataSourceTable,
)
from .query_data_source_table.query_data_source_table import QueryDataSourceTable
from .query_multiple_columns import QueryMultipleColumns
from .query_row_count.query_row_count import QueryRowCount
from .query_row_count.unexpected_rows_query_row_count import UnexpectedRowsQueryRowCount
from .query_table.base_query_table import BaseQueryTable
from .query_table.query_table import QueryTable
from .query_table.unexpected_rows_query_table import UnexpectedRowsQueryTable
from .query_template_values import QueryTemplateValues



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Literal, Union

from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.fluent_base_model import FluentBaseModel

if TYPE_CHECKING:
    from great_expectations.datasource.fluent import BatchParameters
    from great_expectations.datasource.fluent.interfaces import Batch


class _PartitionerDatetime(FluentBaseModel):
    column_name: str
    method_name: str
    sort_ascending: bool = True

    @property
    def columns(self) -> list[str]:
        return [self.column_name]

    def batch_parameters_to_batch_spec_kwarg_identifiers(
        self, parameters: BatchParameters
    ) -> Dict[str, Dict[str, str]]:
        """Validates all the datetime parameters for this partitioner exist in `parameters`."""
        identifiers: Dict = {}
        for part in self.param_names:
            if part in parameters:
                identifiers[part] = parameters[part]
        return {self.column_name: identifiers}

    def _get_concrete_values_from_batch(self, batch: Batch) -> tuple[int]:
        return tuple(batch.metadata[param] for param in self.param_names)

    @property
    def param_names(self) -> list[str]:
        raise NotImplementedError

    def partitioner_method_kwargs(self) -> Dict[str, str]:
        raise NotImplementedError


class DataframePartitionerYearly(_PartitionerDatetime):
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_year"] = "partition_on_year"

    @property
    @override
    def param_names(self) -> List[str]:
        return ["year"]

    @override
    def partitioner_method_kwargs(self) -> Dict[str, str]:
        return {"column_name": self.column_name}


class DataframePartitionerMonthly(_PartitionerDatetime):
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_year_and_month"] = "partition_on_year_and_month"

    @property
    @override
    def param_names(self) -> List[str]:
        return ["year", "month"]

    @override
    def partitioner_method_kwargs(self) -> Dict[str, str]:
        return {"column_name": self.column_name}


class DataframePartitionerDaily(_PartitionerDatetime):
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_year_and_month_and_day"] = (
        "partition_on_year_and_month_and_day"
    )

    @property
    @override
    def param_names(self) -> List[str]:
        return ["year", "month", "day"]

    @override
    def partitioner_method_kwargs(self) -> Dict[str, str]:
        return {"column_name": self.column_name}


DataframePartitioner = Union[
    DataframePartitionerDaily, DataframePartitionerMonthly, DataframePartitionerYearly
]



import pandas as pd
import pytest

from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.metrics.column.aggregate_non_null_count import (
    ColumnAggregateNonNullCount,
    ColumnAggregateNonNullCountResult,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.metrics.conftest import (
    PANDAS_DATA_SOURCES,
    SPARK_DATA_SOURCES,
    SQL_DATA_SOURCES,
)

STRING_COLUMN_NAME = "whatevs"
DATA_FRAME = pd.DataFrame(
    {
        STRING_COLUMN_NAME: ["a", None, "c", "d", None],
    },
    dtype="object",
)

ALL_NULL_DATA_FRAME = pd.DataFrame(
    {
        STRING_COLUMN_NAME: [None, None, None],
    },
    dtype="object",
)

NO_NULL_DATA_FRAME = pd.DataFrame(
    {
        STRING_COLUMN_NAME: ["a", "b", "c", "d"],
    },
    dtype="object",
)

# Empty dataframe with explicit schema for Spark compatibility
EMPTY_DATA_FRAME = pd.DataFrame(
    {
        STRING_COLUMN_NAME: pd.Series([], dtype="object"),
    }
)


class TestColumnAggregateNonNullCount:
    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES + PANDAS_DATA_SOURCES + SPARK_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success(self, batch_for_datasource: Batch) -> None:
        metric = ColumnAggregateNonNullCount(column=STRING_COLUMN_NAME)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnAggregateNonNullCountResult)
        assert metric_result.value == 3

    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES + PANDAS_DATA_SOURCES,
        data=ALL_NULL_DATA_FRAME,
    )
    def test_all_null(self, batch_for_datasource: Batch) -> None:
        metric = ColumnAggregateNonNullCount(column=STRING_COLUMN_NAME)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnAggregateNonNullCountResult)
        assert metric_result.value == 0

    @parameterize_batch_for_data_sources(
        data_source_configs=SPARK_DATA_SOURCES,
        data=ALL_NULL_DATA_FRAME,
    )
    @pytest.mark.xfail(reason="Spark cannot determine types from all-null dataset", strict=True)
    def test_all_null_spark(self, batch_for_datasource: Batch) -> None:
        metric = ColumnAggregateNonNullCount(column=STRING_COLUMN_NAME)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnAggregateNonNullCountResult)
        assert metric_result.value == 0

    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES + PANDAS_DATA_SOURCES + SPARK_DATA_SOURCES,
        data=NO_NULL_DATA_FRAME,
    )
    def test_no_null(self, batch_for_datasource: Batch) -> None:
        metric = ColumnAggregateNonNullCount(column=STRING_COLUMN_NAME)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnAggregateNonNullCountResult)
        assert metric_result.value == 4

    # Simplified empty dataset test - only using Pandas data sources
    @parameterize_batch_for_data_sources(
        data_source_configs=PANDAS_DATA_SOURCES,
        data=EMPTY_DATA_FRAME,
    )
    def test_empty_dataset(self, batch_for_datasource: Batch) -> None:
        """Test the metric with an empty dataset using Pandas data sources."""
        metric = ColumnAggregateNonNullCount(column=STRING_COLUMN_NAME)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnAggregateNonNullCountResult)
        assert metric_result.value == 0



from unittest import mock

import pytest
import pytest_mock

from great_expectations.checkpoint.checkpoint import CheckpointResult
from great_expectations.core.expectation_validation_result import ExpectationSuiteValidationResult
from great_expectations.data_context.data_context.cloud_data_context import (
    CloudDataContext,
)


@pytest.mark.cloud
def test_view_validation_result(
    empty_cloud_data_context: CloudDataContext,
    mocker: pytest_mock.MockFixture,
):
    context = empty_cloud_data_context

    result_url_1 = "https://my.cloud.app/validation-result/123"
    result_url_2 = "https://my.cloud.app/validation-result/123"
    validation_result_1 = mocker.Mock(
        spec=ExpectationSuiteValidationResult, result_url=result_url_1
    )
    validation_result_2 = mocker.Mock(
        spec=ExpectationSuiteValidationResult, result_url=result_url_2
    )
    checkpoint_result = mocker.Mock(
        spec=CheckpointResult,
        run_results={
            "validation_result_1": validation_result_1,
            "validation_result_2": validation_result_2,
        },
    )

    with mock.patch("webbrowser.open") as mock_open:
        context.view_validation_result(checkpoint_result)

    assert mock_open.call_count == 2
    mock_open.assert_any_call(result_url_1)
    mock_open.assert_any_call(result_url_2)



import pytest

import great_expectations as gx
from great_expectations.core import (
    ExpectationSuite,
)
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.render import (
    RenderedDocumentContent,
    RenderedHeaderContent,
    RenderedSectionContent,
    RenderedTableContent,
)
from great_expectations.render.renderer import (
    ExpectationSuitePageRenderer,
)
from great_expectations.render.view import DefaultMarkdownPageView

# module level markers
pytestmark = pytest.mark.big


@pytest.fixture()
def expectation_suite_to_render_with_notes():
    expectation_suite = ExpectationSuite(
        name="default",
        meta={"great_expectations_version": "0.13.0-test"},
        expectations=[
            ExpectationConfiguration(
                type="expect_column_to_exist",
                kwargs={"column": "infinities"},
            ),
            ExpectationConfiguration(type="expect_column_to_exist", kwargs={"column": "nulls"}),
            ExpectationConfiguration(type="expect_column_to_exist", kwargs={"column": "naturals"}),
            ExpectationConfiguration(
                type="expect_column_distinct_values_to_be_in_set",
                kwargs={"column": "irrationals", "value_set": ["*", "1st", "2nd"]},
            ),
            ExpectationConfiguration(
                type="expect_column_values_to_be_unique",
                kwargs={"column": "testings"},
                notes=[
                    "Example notes about this expectation. **Markdown** `Supported`.",
                    "Second example note **with** *Markdown*",
                ],
            ),
        ],
    )
    return expectation_suite


def test_render_section_page():
    section = RenderedSectionContent(
        **{
            "section_name": None,
            "content_blocks": [
                RenderedHeaderContent(
                    **{
                        "content_block_type": "header",
                        "header": "Overview",
                    }
                ),
                RenderedTableContent(
                    **{
                        "content_block_type": "table",
                        "header": "Dataset info",
                        "table": [
                            ["Number of variables", "12"],
                            ["Number of observations", "891"],
                        ],
                        "styling": {
                            "classes": ["col-6", "table-responsive"],
                            "styles": {"margin-top": "20px"},
                            "body": {"classes": ["table", "table-sm"]},
                        },
                    }
                ),
            ],
        }
    )

    rendered_doc = gx.render.view.view.DefaultMarkdownPageView().render(
        RenderedDocumentContent(sections=[section])
    )

    rendered_doc = rendered_doc.replace(" ", "").replace("\t", "").replace("\n", "")

    assert (
        rendered_doc
        == """
        #ValidationResults
        ##Overview
        ###Datasetinfo
        ||||------------|------------|Numberofvariables|12Numberofobservations|891
        -----------------------------------------------------------
        Poweredby[GreatExpectations](https://greatexpectations.io/)
        """.replace(" ", "")
        .replace("\t", "")
        .replace("\n", "")
    )


def test_render_expectation_suite_for_Markdown(expectation_suite_to_render_with_notes):
    expectation_suite_page_renderer = ExpectationSuitePageRenderer()
    rendered_document_content_list = expectation_suite_page_renderer.render(
        expectation_suite_to_render_with_notes
    )
    md_str = DefaultMarkdownPageView().render(rendered_document_content_list)
    md_str = " ".join(md_str)
    md_str = md_str.replace(" ", "").replace("\t", "").replace("\n", "")
    assert (
        md_str
        == r"""
   # Validation Results
## Overview
### Info
 |  |  |
 | ------------  | ------------ |
Expectation Suite Name  | default
Great Expectations Version  | 0.13.0-test
### Notes
    This Expectation suite currently contains 5 total Expectations across 5 columns.
## infinities
  * is a required field.
  * ***
## irrationals
  * distinct values must belong to this set: \* **1st** **2nd**.
  * ***
## naturals
  * is a required field.
  * ***
## nulls
  * is a required field.
  * ***
## testings
  * values must be unique.
    #### Notes:
      Example notes about this expectation. **Markdown** `Supported`.

      Second example note **with** *Markdown*
  * ***
-----------------------------------------------------------
Powered by [Great Expectations](https://greatexpectations.io/)
    """.replace(" ", "")
        .replace("\t", "")
        .replace("\n", "")
    )



from typing import Mapping, Optional

import pandas as pd
import pytest

from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.datasource.fluent.pandas_datasource import DataFrameAsset
from tests.integration.sql_session_manager import SessionSQLEngineManager
from tests.integration.test_utils.data_source_config.base import (
    BatchTestSetup,
    DataSourceTestConfig,
)


class PandasDataFrameDatasourceTestConfig(DataSourceTestConfig):
    @property
    @override
    def label(self) -> str:
        return "pandas-data-frame"

    @property
    @override
    def pytest_mark(self) -> pytest.MarkDecorator:
        return pytest.mark.unit

    @override
    def create_batch_setup(
        self,
        request: pytest.FixtureRequest,
        data: pd.DataFrame,
        extra_data: Mapping[str, pd.DataFrame],
        context: AbstractDataContext,
        engine_manager: Optional[SessionSQLEngineManager] = None,
    ) -> BatchTestSetup:
        assert not extra_data, "extra_data is not supported for this data source."
        return PandasDataFrameBatchTestSetup(data=data, config=self, context=context)


class PandasDataFrameBatchTestSetup(
    BatchTestSetup[PandasDataFrameDatasourceTestConfig, DataFrameAsset]
):
    @override
    def make_asset(self) -> DataFrameAsset:
        return self.context.data_sources.add_pandas(
            self._random_resource_name()
        ).add_dataframe_asset(self._random_resource_name())

    @override
    def make_batch(self) -> Batch:
        return (
            self.make_asset()
            .add_batch_definition_whole_dataframe(self._random_resource_name())
            .get_batch(batch_parameters={"dataframe": self.data})
        )

    @override
    def setup(self) -> None: ...

    @override
    def teardown(self) -> None: ...



from typing import Iterable, Optional, Union

import pytest

from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.execution_engine import SqlAlchemyExecutionEngine
from great_expectations.execution_engine.sqlalchemy_batch_data import SqlAlchemyBatchData


class Dialect:
    def __init__(self, dialect: str):
        self.name = dialect


class MockSaEngine:
    def __init__(self, dialect: Dialect):
        self.dialect = dialect

    def connect(self) -> None:
        pass


class MockResult:
    def fetchmany(self, recordcount: int):
        return None


class MockConnection:
    def execute(self, query: str):
        return MockResult()


_batch_selectable = sa.Table("my_table", sa.MetaData(), schema=None)


@pytest.fixture
def batch_selectable() -> sa.Table:
    return _batch_selectable


class MockSqlAlchemyExecutionEngine(SqlAlchemyExecutionEngine):
    def __init__(self, create_temp_table: bool = True, *args, **kwargs):
        self.engine = MockSaEngine(dialect=Dialect("sqlite"))  # type: ignore[assignment] # FIXME CoP
        self._create_temp_table = create_temp_table
        self._connection = MockConnection()

        self._batch_manager = None

    def get_compute_domain(
        self,
        domain_kwargs: dict,
        domain_type: Union[str, MetricDomainTypes],
        accessor_keys: Optional[Iterable[str]] = None,
    ) -> tuple[sa.Table, dict, dict]:
        return _batch_selectable, {}, {}


class MockBatchManager:
    active_batch_data = SqlAlchemyBatchData(
        execution_engine=MockSqlAlchemyExecutionEngine(),
        table_name="my_table",
    )

    def save_batch_data(self) -> None: ...


@pytest.fixture
def mock_sqlalchemy_execution_engine():
    execution_engine = MockSqlAlchemyExecutionEngine()
    execution_engine._batch_manager = MockBatchManager()
    return execution_engine



import pytest
from marshmallow import Schema

from great_expectations.core.configuration import AbstractConfigSchema
from great_expectations.core.serializer import (
    DictConfigSerializer,
    JsonConfigSerializer,
)


@pytest.fixture
def empty_abstract_config_schema():
    class EmptyAbstractConfigSchema(AbstractConfigSchema):
        pass

    schema = EmptyAbstractConfigSchema()
    return schema


@pytest.mark.unit
def test_init_dict_config_serializer(empty_abstract_config_schema: Schema):
    serializer = DictConfigSerializer(schema=empty_abstract_config_schema)

    assert serializer.schema == empty_abstract_config_schema


@pytest.mark.unit
def test_init_json_config_serializer(empty_abstract_config_schema: Schema):
    serializer = JsonConfigSerializer(schema=empty_abstract_config_schema)

    assert serializer.schema == empty_abstract_config_schema



"""
Test GXCloudStoreBackend behavior and adherence to StoreBackend contract.

Since GXCloudStoreBackend relies on GX Cloud, we mock requests and assert the right calls
are made from the Store API (set, get, list, and remove_key).

Note that although ge_cloud_access_token is provided (and is a valid UUID), no external
requests are actually made as part of this test. The actual value of the token does not
matter here but we leverage an existing fixture to mimic the contents of requests made
in production. The same logic applies to all UUIDs in this test.
"""

from typing import Callable, Optional, Set, Union
from unittest import mock

import pytest
import responses

import great_expectations.exceptions as gx_exceptions
from great_expectations.data_context.cloud_constants import (
    CLOUD_DEFAULT_BASE_URL,
    GXCloudRESTResource,
)
from great_expectations.data_context.store.gx_cloud_store_backend import (
    GXCloudStoreBackend,
)

# module level markers
pytestmark = pytest.mark.cloud


@pytest.fixture
def construct_ge_cloud_store_backend(
    ge_cloud_access_token: str,
) -> Callable[[GXCloudRESTResource], GXCloudStoreBackend]:
    def _closure(resource_type: GXCloudRESTResource) -> GXCloudStoreBackend:
        ge_cloud_base_url = CLOUD_DEFAULT_BASE_URL
        ge_cloud_credentials = {
            "access_token": ge_cloud_access_token,
            "organization_id": "51379b8b-86d3-4fe7-84e9-e1a52f4a414c",
        }

        store_backend = GXCloudStoreBackend(
            ge_cloud_base_url=ge_cloud_base_url,
            ge_cloud_credentials=ge_cloud_credentials,
            ge_cloud_resource_type=resource_type,
        )
        return store_backend

    return _closure


@pytest.mark.parametrize(
    "base_url,organization_id,resource_name,id,expected",
    [
        pytest.param(
            "https://app.test.greatexpectations.io",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "my_resource",
            None,
            "https://app.test.greatexpectations.io/organizations/de5b9ca6-caf7-43c8-a820-5540ec6df9b2/my-resource",
            id="no id",
        ),
        pytest.param(
            "https://app.test.greatexpectations.io",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "my_resource",
            "8746e25c-de4d-450d-967b-df0d5546590d",
            "https://app.test.greatexpectations.io/organizations/de5b9ca6-caf7-43c8-a820-5540ec6df9b2/my-resource/8746e25c-de4d-450d-967b-df0d5546590d",
            id="with id",
        ),
        pytest.param(
            "https://app.test.greatexpectations.io",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "expectation_suites",
            "8746e25c-de4d-450d-967b-df0d5546590d",
            "https://app.test.greatexpectations.io/api/v1/organizations/de5b9ca6-caf7-43c8-a820-5540ec6df9b2/expectation-suites/8746e25c-de4d-450d-967b-df0d5546590d",
            id="expectation-suites V1",
        ),
    ],
)
def test_construct_url(
    base_url: str,
    organization_id: str,
    resource_name: str,
    id: Optional[str],
    expected: str,
) -> None:
    assert (
        GXCloudStoreBackend.construct_versioned_url(
            base_url=base_url,
            organization_id=organization_id,
            resource_name=resource_name,
            id=id,
        )
        == expected
    )


@pytest.mark.parametrize(
    "resource_type,organization_id,attributes_key,attributes_value,kwargs,expected",
    [
        pytest.param(
            "my_resource",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "my_attribute",
            "my_value",
            {},
            {
                "data": {
                    "type": "my_resource",
                    "attributes": {
                        "organization_id": "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
                        "my_attribute": "my_value",
                    },
                },
            },
            id="no kwargs",
        ),
        pytest.param(
            "my_resource",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "my_attribute",
            {"key1": {"nested_key1": 1}, "key2": {"nested_key2": 2}},
            {},
            {
                "data": {
                    "type": "my_resource",
                    "attributes": {
                        "organization_id": "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
                        "my_attribute": {
                            "key1": {"nested_key1": 1},
                            "key2": {"nested_key2": 2},
                        },
                    },
                },
            },
            id="with nested value",
        ),
        pytest.param(
            "my_resource",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "my_attribute",
            "my_value",
            {
                "kwarg1": 1,
                "kwarg2": 2,
            },
            {
                "data": {
                    "type": "my_resource",
                    "attributes": {
                        "organization_id": "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
                        "my_attribute": "my_value",
                        "kwarg1": 1,
                        "kwarg2": 2,
                    },
                },
            },
            id="with basic kwargs",
        ),
        pytest.param(
            "my_resource",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "my_attribute",
            "my_value",
            {
                "kwarg1": {
                    "nested1": 1,
                },
                "kwarg2": {
                    "nested2": 2,
                },
            },
            {
                "data": {
                    "type": "my_resource",
                    "attributes": {
                        "organization_id": "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
                        "my_attribute": "my_value",
                        "kwarg1": {
                            "nested1": 1,
                        },
                        "kwarg2": {
                            "nested2": 2,
                        },
                    },
                },
            },
            id="with nested kwargs",
        ),
        pytest.param(
            "expectation_suite",
            "de5b9ca6-caf7-43c8-a820-5540ec6df9b2",
            "v1_configs_dont_use_attribute_key",
            {"expectations": [], "meta": None, "notes": None},
            {},
            {
                "data": {"expectations": [], "meta": None, "notes": None},
            },
            id="V1 expectation suite",
        ),
    ],
)
def test_construct_json_payload(
    resource_type: str,
    organization_id: str,
    attributes_key: str,
    attributes_value: str,
    kwargs: dict,
    expected: dict,
) -> None:
    assert (
        GXCloudStoreBackend.construct_versioned_payload(
            resource_type=resource_type,
            organization_id=organization_id,
            attributes_key=attributes_key,
            attributes_value=attributes_value,
            **kwargs,
        )
        == expected
    )


@pytest.mark.cloud
def test_construct_json_payload_raises_error_with_V1_resource_and_wrong_attributes_value_type():
    v1_resource = "expectation_suite"
    organization_id = "de5b9ca6-caf7-43c8-a820-5540ec6df9b2"
    attributes_value_of_legacy_type = "a string"
    with pytest.raises(
        TypeError,
        match="Parameter attributes_value of type <class 'str'> is unsupported in GX V1.",
    ):
        GXCloudStoreBackend.construct_versioned_payload(
            resource_type=v1_resource,
            organization_id=organization_id,
            attributes_key="",
            attributes_value=attributes_value_of_legacy_type,
        )


def test_list_keys(
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.CHECKPOINT)

    with mock.patch("requests.Session.get", autospec=True) as mock_get:
        store_backend.list_keys()
        mock_get.assert_called_with(
            mock.ANY,  # requests.Session object
            url=f"{CLOUD_DEFAULT_BASE_URL}api/v1/organizations/51379b8b-86d3-4fe7-84e9-e1a52f4a414c/checkpoints",
            params=None,
        )


@responses.activate
def test_list_keys_with_empty_payload_from_backend(
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
):
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.DATASOURCE)

    responses.add(
        responses.GET,
        f"{CLOUD_DEFAULT_BASE_URL}api/v1/organizations/51379b8b-86d3-4fe7-84e9-e1a52f4a414c/datasources",
        json={"data": []},
        status=200,
    )

    assert store_backend.list_keys() == []
    assert len(responses.calls) == 1


@responses.activate
def test_has_key_with_empty_payload_from_backend(
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
):
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.EXPECTATION_SUITE)

    name = "my_nonexistent_suite"
    responses.add(
        responses.GET,
        f"{CLOUD_DEFAULT_BASE_URL}api/v1/organizations/51379b8b-86d3-4fe7-84e9-e1a52f4a414c/expectation-suites?name={name}",
        json={"data": []},
        status=200,
    )

    key = (GXCloudRESTResource.EXPECTATION_SUITE, None, name)
    assert store_backend.has_key(key) is False
    assert len(responses.calls) == 1


@responses.activate
def test_get_with_empty_payload_from_backend(
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
):
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.EXPECTATION_SUITE)

    name = "my_nonexistent_suite"
    responses.add(
        responses.GET,
        f"{CLOUD_DEFAULT_BASE_URL}api/v1/organizations/51379b8b-86d3-4fe7-84e9-e1a52f4a414c/expectation-suites?name={name}",
        json={"data": []},
        status=200,
    )

    key = (GXCloudRESTResource.EXPECTATION_SUITE, None, name)

    with pytest.raises(gx_exceptions.StoreBackendError):
        _ = store_backend.get(key)
    assert len(responses.calls) == 1


def test_get_all(
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.DATASOURCE)

    with mock.patch("requests.Session.get", autospec=True) as mock_get:
        store_backend.get_all()
        mock_get.assert_called_with(
            mock.ANY,  # requests.Session object
            url=f"{CLOUD_DEFAULT_BASE_URL}api/v1/organizations/51379b8b-86d3-4fe7-84e9-e1a52f4a414c/datasources",
            params=None,
        )


def test_remove_key_with_only_id(
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.CHECKPOINT)

    with mock.patch("requests.Session.delete", autospec=True) as mock_delete:
        mock_response = mock_delete.return_value
        mock_response.status_code = 204

        store_backend.remove_key(
            (
                "checkpoint",
                "0ccac18e-7631-4bdd-8a42-3c35cce574c6",
            )
        )
        mock_delete.assert_called_with(
            mock.ANY,  # requests.Session object
            f"{CLOUD_DEFAULT_BASE_URL}api/v1/organizations/51379b8b-86d3-4fe7-84e9-e1a52f4a414c/checkpoints/0ccac18e-7631"
            "-4bdd"
            "-8a42-3c35cce574c6",
        )


def test_remove_key_with_id_and_name(
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.CHECKPOINT)

    with mock.patch("requests.Session.delete", autospec=True) as mock_delete:
        mock_response = mock_delete.return_value
        mock_response.status_code = 204

        store_backend.remove_key(
            ("checkpoint", "0ccac18e-7631-4bdd-8a42-3c35cce574c6", "checkpoint_name")
        )
        mock_delete.assert_called_with(
            mock.ANY,  # requests.Session object
            f"{CLOUD_DEFAULT_BASE_URL}api/v1/organizations/51379b8b-86d3-4fe7-84e9-e1a52f4a414c/checkpoints/0ccac18e-7631"
            "-4bdd"
            "-8a42-3c35cce574c6",
        )


def test_remove_key_with_only_name(
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.CHECKPOINT)

    with mock.patch("requests.Session.delete", autospec=True) as mock_delete:
        mock_response = mock_delete.return_value
        mock_response.status_code = 200

        store_backend.remove_key(("checkpoint", "", "checkpoint_name"))
        mock_delete.assert_called_with(
            mock.ANY,  # requests.Session object
            f"{CLOUD_DEFAULT_BASE_URL}api/v1/organizations/51379b8b-86d3-4fe7-84e9-e1a52f4a414c/checkpoints",
            params={"name": "checkpoint_name"},
        )


def test_appropriate_casting_of_str_resource_type_to_GXCloudRESTResource(
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.CHECKPOINT)

    assert store_backend.ge_cloud_resource_type is GXCloudRESTResource.CHECKPOINT


@pytest.mark.parametrize(
    "resource_type,expected_set_kwargs",
    [
        pytest.param(
            GXCloudRESTResource.VALIDATION_RESULT,
            {"checkpoint_id", "expectation_suite_id"},
            id="resource_with_allowed_set_kwargs",
        ),
        pytest.param(
            GXCloudRESTResource.CHECKPOINT,
            set(),
            id="resource_without_allowed_set_kwargs",
        ),
    ],
)
def test_allowed_set_kwargs(
    resource_type: GXCloudRESTResource,
    expected_set_kwargs: Set[str],
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
) -> None:
    store_backend = construct_ge_cloud_store_backend(resource_type)
    actual = store_backend.allowed_set_kwargs

    assert actual == expected_set_kwargs


@pytest.mark.parametrize(
    "kwargs,expected",
    [
        pytest.param(
            {},
            True,
            id="no kwargs",
        ),
        pytest.param(
            {"checkpoint_id": "123", "expectation_suite_id": "456"},
            True,
            id="equivalent kwargs",
        ),
        pytest.param(
            {
                "checkpoint_id": "123",
                "expectation_suite_id": "456",
                "my_other_id": "789",
            },
            None,
            id="additional kwargs",
            marks=pytest.mark.xfail(strict=True, raises=ValueError),
        ),
    ],
)
def test_validate_set_kwargs(
    kwargs: dict,
    expected: Union[bool, None],
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.VALIDATION_RESULT)
    assert store_backend.validate_set_kwargs(kwargs) == expected


def test_config_property_and_defaults(
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
) -> None:
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.CHECKPOINT)

    assert store_backend.config == {
        "class_name": GXCloudStoreBackend.__name__,
        "fixed_length_key": True,
        "ge_cloud_base_url": CLOUD_DEFAULT_BASE_URL,
        "ge_cloud_resource_type": GXCloudRESTResource.CHECKPOINT,
        "manually_initialize_store_backend_id": "",
        "module_name": GXCloudStoreBackend.__module__,
        "suppress_store_backend_id": True,
    }


@pytest.mark.parametrize(
    "key",
    [
        pytest.param(("checkpoint", None), id="invalid length"),
        pytest.param(
            (
                "my_fake_gx_resource",
                "9805c644-6728-459f-846e-db7b40e8e811",
                "test_resource",
            ),
            id="invalid resource type",
        ),
    ],
)
def test_ge_cloud_store_backend_invalid_key_raises_error(
    ge_cloud_access_token: str,
    key: tuple,
) -> None:
    ge_cloud_credentials = {
        "access_token": ge_cloud_access_token,
        "organization_id": "51379b8b-86d3-4fe7-84e9-e1a52f4a414c",
    }

    backend = GXCloudStoreBackend(
        ge_cloud_credentials=ge_cloud_credentials,
        ge_cloud_resource_type=GXCloudRESTResource.CHECKPOINT,
    )

    with pytest.raises(TypeError):
        backend.get(key)


def test_closes_session(
    construct_ge_cloud_store_backend: Callable[[GXCloudRESTResource], GXCloudStoreBackend],
) -> None:
    """Make sure that when the GXCloudStoreBackend object is garbage collected,
    the session is closed."""
    store_backend = construct_ge_cloud_store_backend(GXCloudRESTResource.CHECKPOINT)

    with mock.patch("requests.Session.close", autospec=True) as mock_close:
        # Use the finalizer to remove the object from memory.
        # Using del or __del__ directly does not work in the test environment.
        store_backend._finalizer()
        mock_close.assert_called_once()



from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import ClassVar, List, Literal

from great_expectations.analytics.actions import (
    CHECKPOINT_CREATED,
    CHECKPOINT_DELETED,
    CHECKPOINT_RAN,
    DATA_CONTEXT_INITIALIZED,
    DOMAIN_OBJECT_ALL_DESERIALIZE_ERROR,
    EXPECTATION_SUITE_CREATED,
    EXPECTATION_SUITE_DELETED,
    EXPECTATION_SUITE_EXPECTATION_CREATED,
    EXPECTATION_SUITE_EXPECTATION_DELETED,
    EXPECTATION_SUITE_EXPECTATION_UPDATED,
    NOTIFICATION_ACTION_RAN,
    VALIDATION_DEFINITION_CREATED,
    VALIDATION_DEFINITION_DELETED,
    VALIDATION_DEFINITION_RAN,
)
from great_expectations.analytics.base_event import Action, Event
from great_expectations.compatibility.typing_extensions import override


@dataclass
class DataContextInitializedEvent(Event):
    _allowed_actions: ClassVar[List[Action]] = [DATA_CONTEXT_INITIALIZED]

    def __init__(self):
        super().__init__(action=DATA_CONTEXT_INITIALIZED)


@dataclass
class _ExpectationSuiteExpectationEvent(Event):
    expectation_id: str | None = None
    expectation_suite_id: str | None = None

    @override
    def _properties(self) -> dict:
        return {
            "expectation_id": self.expectation_id,
            "expectation_suite_id": self.expectation_suite_id,
        }


@dataclass
class ExpectationSuiteExpectationCreatedEvent(_ExpectationSuiteExpectationEvent):
    expectation_type: str = "UNKNOWN"
    custom_exp_type: bool = False

    _allowed_actions: ClassVar[List[Action]] = [
        EXPECTATION_SUITE_EXPECTATION_CREATED,
    ]

    def __init__(
        self,
        expectation_id: str | None = None,
        expectation_suite_id: str | None = None,
        expectation_type: str = "UNKNOWN",
        custom_exp_type: bool = False,
    ):
        super().__init__(
            action=EXPECTATION_SUITE_EXPECTATION_CREATED,
            expectation_id=expectation_id,
            expectation_suite_id=expectation_suite_id,
        )
        self.expectation_type = expectation_type
        self.custom_exp_type = custom_exp_type

    @override
    def _properties(self) -> dict:
        return {
            **super()._properties(),
            "expectation_type": self.expectation_type,
            "custom_exp_type": self.custom_exp_type,
        }


@dataclass
class ExpectationSuiteExpectationUpdatedEvent(_ExpectationSuiteExpectationEvent):
    _allowed_actions: ClassVar[List[Action]] = [
        EXPECTATION_SUITE_EXPECTATION_UPDATED,
    ]

    def __init__(
        self,
        expectation_id: str | None = None,
        expectation_suite_id: str | None = None,
    ):
        super().__init__(
            action=EXPECTATION_SUITE_EXPECTATION_UPDATED,
            expectation_id=expectation_id,
            expectation_suite_id=expectation_suite_id,
        )


@dataclass
class ExpectationSuiteExpectationDeletedEvent(_ExpectationSuiteExpectationEvent):
    _allowed_actions: ClassVar[List[Action]] = [
        EXPECTATION_SUITE_EXPECTATION_DELETED,
    ]

    def __init__(
        self,
        expectation_id: str | None = None,
        expectation_suite_id: str | None = None,
    ):
        super().__init__(
            action=EXPECTATION_SUITE_EXPECTATION_DELETED,
            expectation_id=expectation_id,
            expectation_suite_id=expectation_suite_id,
        )


@dataclass
class _ExpectationSuiteEvent(Event):
    expectation_suite_id: str | None = None

    @override
    def _properties(self) -> dict:
        return {
            "expectation_suite_id": self.expectation_suite_id,
        }


@dataclass
class ExpectationSuiteCreatedEvent(_ExpectationSuiteEvent):
    _allowed_actions: ClassVar[List[Action]] = [EXPECTATION_SUITE_CREATED]

    def __init__(self, expectation_suite_id: str | None = None):
        super().__init__(
            action=EXPECTATION_SUITE_CREATED,
            expectation_suite_id=expectation_suite_id,
        )


@dataclass
class ExpectationSuiteDeletedEvent(_ExpectationSuiteEvent):
    _allowed_actions: ClassVar[List[Action]] = [EXPECTATION_SUITE_DELETED]

    def __init__(self, expectation_suite_id: str | None = None):
        super().__init__(
            action=EXPECTATION_SUITE_DELETED,
            expectation_suite_id=expectation_suite_id,
        )


@dataclass
class _CheckpointEvent(Event):
    checkpoint_id: str | None = None

    @override
    def _properties(self) -> dict:
        return {
            "checkpoint_id": self.checkpoint_id,
        }


@dataclass
class ActionInfo:
    type: str
    notify_on: Literal["all", "failure", "success"] | None


@dataclass
class CheckpointCreatedEvent(_CheckpointEvent):
    _allowed_actions: ClassVar[List[Action]] = [CHECKPOINT_CREATED]

    def __init__(
        self,
        checkpoint_id: str | None = None,
        validation_definition_ids: list[str | None] | None = None,
        actions: list[ActionInfo] | None = None,
    ):
        self.validation_definition_ids = validation_definition_ids
        self.actions = actions or []
        super().__init__(
            action=CHECKPOINT_CREATED,
            checkpoint_id=checkpoint_id,
        )

    @override
    def _properties(self) -> dict:
        return {
            "validation_definition_ids": self.validation_definition_ids,
            "actions": [asdict(action) for action in self.actions],
            **super()._properties(),
        }


@dataclass
class CheckpointDeletedEvent(_CheckpointEvent):
    _allowed_actions: ClassVar[List[Action]] = [CHECKPOINT_DELETED]

    def __init__(self, checkpoint_id: str | None = None):
        super().__init__(
            action=CHECKPOINT_DELETED,
            checkpoint_id=checkpoint_id,
        )


@dataclass
class CheckpointRanEvent(_CheckpointEvent):
    _allowed_actions: ClassVar[List[Action]] = [CHECKPOINT_RAN]

    def __init__(
        self,
        checkpoint_id: str | None = None,
        validation_definition_ids: list[str | None] | None = None,
    ):
        self.validation_definition_ids = validation_definition_ids
        super().__init__(
            action=CHECKPOINT_RAN,
            checkpoint_id=checkpoint_id,
        )

    @override
    def _properties(self) -> dict:
        return {
            "validation_definition_ids": self.validation_definition_ids,
            **super()._properties(),
        }


@dataclass
class _ValidationDefinitionEvent(Event):
    validation_definition_id: str | None = None

    @override
    def _properties(self) -> dict:
        return {
            "validation_definition_id": self.validation_definition_id,
        }


@dataclass
class ValidationDefinitionCreatedEvent(_ValidationDefinitionEvent):
    _allowed_actions: ClassVar[List[Action]] = [VALIDATION_DEFINITION_CREATED]

    def __init__(self, validation_definition_id: str | None = None):
        super().__init__(
            action=VALIDATION_DEFINITION_CREATED,
            validation_definition_id=validation_definition_id,
        )


@dataclass
class ValidationDefinitionDeletedEvent(_ValidationDefinitionEvent):
    _allowed_actions: ClassVar[List[Action]] = [VALIDATION_DEFINITION_DELETED]

    def __init__(self, validation_definition_id: str | None = None):
        super().__init__(
            action=VALIDATION_DEFINITION_DELETED,
            validation_definition_id=validation_definition_id,
        )


@dataclass
class ValidationDefinitionRanEvent(_ValidationDefinitionEvent):
    checkpoint_id: str | None = None
    _allowed_actions: ClassVar[List[Action]] = [VALIDATION_DEFINITION_RAN]

    @override
    def _properties(self) -> dict:
        return {
            "checkpoint_id": self.checkpoint_id,
            "validation_definition_id": self.validation_definition_id,
        }

    def __init__(
        self,
        *,
        validation_definition_id: str | None = None,
        checkpoint_id: str | None = None,
    ):
        self.checkpoint_id = checkpoint_id
        super().__init__(
            action=VALIDATION_DEFINITION_RAN,
            validation_definition_id=validation_definition_id,
        )


@dataclass
class DomainObjectAllDeserializationEvent(Event):
    _allowed_actions: ClassVar[List[Action]] = [DOMAIN_OBJECT_ALL_DESERIALIZE_ERROR]

    store_name: str
    error_type: str

    def __init__(self, store_name: str, error_type: str):
        super().__init__(action=DOMAIN_OBJECT_ALL_DESERIALIZE_ERROR)
        self.error_type = error_type
        self.store_name = store_name

    @override
    def _properties(self) -> dict:
        return {
            "error_type": self.error_type,
            "store_name": self.store_name,
        }


@dataclass
class NotificationActionRanEvent(Event):
    _allowed_actions: ClassVar[List[Action]] = [NOTIFICATION_ACTION_RAN]

    def __init__(
        self,
        type: Literal["slack", "microsoft", "email"],
        notify_type: Literal["all", "failure", "success"],
        checkpoint_id: str | None,
    ):
        self.type = type
        self.notify_type = notify_type
        self.checkpoint_id = checkpoint_id
        super().__init__(
            action=NOTIFICATION_ACTION_RAN,
        )

    @override
    def _properties(self) -> dict:
        return {
            "type": self.type,
            "checkpoint_id": self.checkpoint_id,
            "notify_type": self.notify_type,
        }



from __future__ import annotations

import functools
import logging
import pathlib
import urllib.parse
import warnings
from contextlib import contextmanager
from pprint import pformat as pf
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ContextManager,
    Dict,
    Final,
    Generator,
    List,
    Optional,
    Type,
    Union,
)

import pytest
from moto import mock_s3
from pytest import MonkeyPatch
from typing_extensions import TypeAlias, override

import great_expectations as gx
from great_expectations.compatibility import aws
from great_expectations.core.batch import BatchData
from great_expectations.core.batch_spec import (
    BatchMarkers,
    SqlAlchemyDatasourceBatchSpec,
)
from great_expectations.data_context import EphemeralDataContext, FileDataContext
from great_expectations.data_context.data_context.abstract_data_context import AbstractDataContext
from great_expectations.datasource.fluent import (
    PandasAzureBlobStorageDatasource,
    PandasGoogleCloudStorageDatasource,
    SparkAzureBlobStorageDatasource,
    SparkGoogleCloudStorageDatasource,
    SQLDatasource,
)
from great_expectations.datasource.fluent.config import GxConfig
from great_expectations.datasource.fluent.interfaces import Datasource
from great_expectations.datasource.fluent.pandas_filesystem_datasource import (
    PandasFilesystemDatasource,
)
from great_expectations.datasource.fluent.postgres_datasource import PostgresDatasource
from great_expectations.datasource.fluent.sources import DataSourceManager
from great_expectations.datasource.fluent.sql_datasource import TableAsset
from great_expectations.execution_engine import (
    ExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from tests.datasource.fluent._fake_cloud_api import (
    _CLOUD_API_FAKE_DB,
    FAKE_ORG_ID,
    GX_CLOUD_MOCK_BASE_URL,
    FakeDBTypedDict,
    create_fake_db_seed_data,
)
from tests.sqlalchemy_test_doubles import Dialect, MockSaEngine

if TYPE_CHECKING:
    import responses
    from botocore.client import BaseClient as BotoBaseClient
    from pytest import FixtureRequest

    from great_expectations.data_context import CloudDataContext


CreateSourceFixture: TypeAlias = Callable[..., ContextManager[PostgresDatasource]]
FLUENT_DATASOURCE_TEST_DIR: Final = pathlib.Path(__file__).parent
PG_CONFIG_YAML_FILE: Final = FLUENT_DATASOURCE_TEST_DIR / FileDataContext.GX_YML
_DEFAULT_TEST_YEARS = list(range(2021, 2023))
_DEFAULT_TEST_MONTHS = list(range(1, 13))


CNF_TEST_LOGGER: Final[logging.Logger] = logging.getLogger(__name__)


def sqlachemy_execution_engine_mock_cls(
    validate_batch_spec: Callable[[SqlAlchemyDatasourceBatchSpec], None],
    dialect: str,
    partitioner_query_response: Optional[Union[List[Dict[str, Any]], List[Any]]] = None,
):
    """Creates a mock gx sql alchemy engine class

    Args:
        validate_batch_spec: A hook that can be used to validate the generated the batch spec
            passed into get_batch_data_and_markers
        dialect: A string representing the SQL Engine dialect. Examples include: postgresql, sqlite
        partitioner_query_response: An optional list of dictionaries. Each dictionary is a row returned
            from the partitioner query. The keys are the column names and the value is the column values,
            eg: [{'year': 2021, 'month': 1}, {'year': 2021, 'month': 2}]
    """  # noqa: E501 # FIXME CoP

    class MockSqlAlchemyExecutionEngine(SqlAlchemyExecutionEngine):
        def __init__(self, create_temp_table: bool = True, *args, **kwargs):
            # We should likely let the user pass in an engine. In a SqlAlchemyExecutionEngine used in  # noqa: E501 # FIXME CoP
            # non-mocked code the engine property is of the type:
            # from sqlalchemy.engine import Engine as SaEngine
            self.engine = MockSaEngine(dialect=Dialect(dialect))  # type: ignore[assignment] # FIXME CoP
            self._create_temp_table = create_temp_table

        @override
        def get_batch_data_and_markers(  # type: ignore[override] # FIXME CoP
            self, batch_spec: SqlAlchemyDatasourceBatchSpec
        ) -> tuple[BatchData, BatchMarkers]:
            validate_batch_spec(batch_spec)
            return BatchData(self), BatchMarkers(ge_load_time=None)

        def execute_partitioned_query(self, partitioned_query):  # type: ignore[explicit-override] # FIXME
            class Row:
                def __init__(self, attributes):
                    for k, v in attributes.items():
                        setattr(self, k, v)

            # We know that partitioner_query_response is non-empty because of validation
            # at the top of the outer function.
            # In some cases, such as in the datetime partitioners,
            # a dictionary is returned our from out partitioner query with the key as the parameter_name.  # noqa: E501 # FIXME CoP
            # Otherwise, a list of values is returned.
            if isinstance(partitioner_query_response[0], dict):
                return [Row(row_dict) for row_dict in partitioner_query_response]
            return partitioner_query_response

    return MockSqlAlchemyExecutionEngine


class ExecutionEngineDouble(ExecutionEngine):
    def __init__(self, *args, **kwargs):
        pass

    @override
    def get_batch_data_and_markers(self, batch_spec) -> tuple[BatchData, BatchMarkers]:  # type: ignore[override] # FIXME CoP
        return BatchData(self), BatchMarkers(ge_load_time=None)


@pytest.fixture
def inject_engine_lookup_double(
    monkeypatch: MonkeyPatch,
) -> Generator[Type[ExecutionEngineDouble], None, None]:
    """
    Inject an execution engine test double into the _SourcesFactory.engine_lookup
    so that all Datasources use the execution engine double.
    Dynamically create a new subclass so that runtime type validation does not fail.
    """
    original_engine_override: dict[Type[Datasource], Type[ExecutionEngine]] = {}
    for key in DataSourceManager.type_lookup:
        if issubclass(type(key), Datasource):
            original_engine_override[key] = key.execution_engine_override

    try:
        for source in original_engine_override:
            source.execution_engine_override = ExecutionEngineDouble
        yield ExecutionEngineDouble
    finally:
        for source, engine in original_engine_override.items():
            source.execution_engine_override = engine


@pytest.fixture
def sqlite_database_path() -> pathlib.Path:
    relative_path = pathlib.Path(
        "..",
        "..",
        "test_sets",
        "taxi_yellow_tripdata_samples",
        "sqlite",
        "yellow_tripdata.db",
    )
    return pathlib.Path(__file__).parent.joinpath(relative_path).resolve(strict=True)


@pytest.fixture
def seed_ds_env_vars(
    monkeypatch: pytest.MonkeyPatch, sqlite_database_path: pathlib.Path
) -> tuple[tuple[str, str], ...]:
    """Seed a collection of ENV variables for use in testing config substitution."""
    config_sub_dict = {
        "MY_CONN_STR": f"sqlite:///{sqlite_database_path}",
        "MY_URL": "http://example.com",
        "MY_FILE": __file__,
    }

    for name, value in config_sub_dict.items():
        monkeypatch.setenv(name, value)
        CNF_TEST_LOGGER.info(f"Setting ENV - {name} = '{value}'")

    # return as tuple of tuples so that the return value is immutable and therefore cacheable
    return tuple((k, v) for k, v in config_sub_dict.items())


@pytest.fixture
def cloud_api_fake_db(cloud_api_fake) -> FakeDBTypedDict:
    from tests.datasource.fluent._fake_cloud_api import _CLOUD_API_FAKE_DB

    return _CLOUD_API_FAKE_DB


@pytest.fixture
def file_dc_config_dir_init(tmp_path: pathlib.Path) -> pathlib.Path:
    """
    Initialize an regular/old-style FileDataContext project config directory.
    Removed on teardown.
    """
    gx_yml = tmp_path / FileDataContext.GX_DIR / FileDataContext.GX_YML
    assert gx_yml.exists() is False
    gx.get_context(mode="file", project_root_dir=tmp_path)
    assert gx_yml.exists()

    tmp_gx_dir = gx_yml.parent.absolute()
    CNF_TEST_LOGGER.info(f"tmp_gx_dir -> {tmp_gx_dir}")
    return tmp_gx_dir


@pytest.fixture
def empty_file_context(file_dc_config_dir_init) -> FileDataContext:
    context = gx.get_context(context_root_dir=file_dc_config_dir_init, cloud_mode=False)
    return context


@pytest.fixture(
    params=[
        pytest.param("empty_cloud_context_fluent", marks=pytest.mark.cloud),
        pytest.param("empty_file_context", marks=pytest.mark.filesystem),
    ],
    ids=["cloud", "file"],
)
def empty_contexts(
    request: FixtureRequest,
    cloud_storage_get_client_doubles,
) -> FileDataContext | CloudDataContext:
    context_fixture: FileDataContext | CloudDataContext = request.getfixturevalue(request.param)
    return context_fixture


@pytest.fixture(scope="session")
def fluent_gx_config_yml() -> pathlib.Path:
    assert PG_CONFIG_YAML_FILE.exists()
    return PG_CONFIG_YAML_FILE


@pytest.fixture(scope="session")
def fluent_gx_config_yml_str(fluent_gx_config_yml: pathlib.Path) -> str:
    return fluent_gx_config_yml.read_text()


@pytest.fixture()
def aws_region_name() -> str:
    return "us-east-1"


@pytest.fixture(scope="function")
def aws_credentials(monkeypatch: pytest.MonkeyPatch) -> None:
    """Monkeypatch ENV AWS Credentials for moto."""
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "testing")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "testing")
    monkeypatch.setenv("AWS_SECURITY_TOKEN", "testing")
    monkeypatch.setenv("AWS_SESSION_TOKEN", "testing")
    monkeypatch.setenv("AWS_DEFAULT_REGION", "testing")


@pytest.fixture
def s3_mock(aws_credentials, aws_region_name: str) -> Generator[BotoBaseClient, None, None]:
    with mock_s3():
        client = aws.boto3.client("s3", region_name=aws_region_name)
        yield client


class _TestClientDummy:
    pass


_CLIENT_DUMMY = _TestClientDummy()


def _get_test_client_dummy(*args, **kwargs) -> _TestClientDummy:
    CNF_TEST_LOGGER.debug(
        f"_get_test_client_dummy() called with \nargs: {pf(args)}\nkwargs: {pf(kwargs)}"
    )
    return _CLIENT_DUMMY


@pytest.fixture
def gcs_get_client_dummy(monkeypatch: MonkeyPatch):
    monkeypatch.setattr(
        PandasGoogleCloudStorageDatasource,
        "_get_gcs_client",
        _get_test_client_dummy,
        raising=True,
    )
    monkeypatch.setattr(
        SparkGoogleCloudStorageDatasource,
        "_get_gcs_client",
        _get_test_client_dummy,
        raising=True,
    )


@pytest.fixture
def azure_get_client_dummy(monkeypatch: MonkeyPatch):
    monkeypatch.setattr(
        PandasAzureBlobStorageDatasource,
        "_get_azure_client",
        _get_test_client_dummy,
        raising=True,
    )
    monkeypatch.setattr(
        SparkAzureBlobStorageDatasource,
        "_get_azure_client",
        _get_test_client_dummy,
        raising=True,
    )


@pytest.fixture
def cloud_storage_get_client_doubles(
    gcs_get_client_dummy,
    azure_get_client_dummy,
):
    """
    Patches Datasources that rely on a private _get_*_client() method to return test doubles
    instead.

    gcs
    azure
    """
    CNF_TEST_LOGGER.warning(
        "Patching cloud storage _get_*_client() methods to return client test doubles"
    )


@pytest.fixture
def filter_data_connector_build_warning():
    with warnings.catch_warnings() as w:
        warnings.simplefilter("ignore", RuntimeWarning)
        yield w


@pytest.fixture
def fluent_only_config(fluent_gx_config_yml_str: str, seed_ds_env_vars: tuple) -> GxConfig:
    """Creates a fluent `GxConfig` object and ensures it contains at least one `Datasource`"""
    fluent_config = GxConfig.parse_yaml(fluent_gx_config_yml_str)
    assert fluent_config.datasources
    return fluent_config


@pytest.fixture
def fluent_yaml_config_file(
    file_dc_config_dir_init: pathlib.Path,
    fluent_gx_config_yml_str: str,
) -> pathlib.Path:
    """
    Dump the provided GxConfig to a temporary path. File is removed during test teardown.

    Append fluent config to default config file
    """
    config_file_path = file_dc_config_dir_init / FileDataContext.GX_YML

    assert config_file_path.exists() is True

    with open(config_file_path, mode="a") as f_append:
        yaml_string = "\n# Fluent\n" + fluent_gx_config_yml_str
        f_append.write(yaml_string)

    CNF_TEST_LOGGER.debug(f"  Config File Text\n-----------\n{config_file_path.read_text()}")
    return config_file_path


@pytest.fixture
@functools.lru_cache(maxsize=1)
def seeded_file_context(
    filter_data_connector_build_warning,
    fluent_yaml_config_file: pathlib.Path,
    seed_ds_env_vars: tuple,
) -> FileDataContext:
    context = gx.get_context(context_root_dir=fluent_yaml_config_file.parent, cloud_mode=False)
    assert isinstance(context, FileDataContext)
    return context


@pytest.fixture
def seed_cloud(
    filter_data_connector_build_warning,
    cloud_api_fake: responses.RequestsMock,
    fluent_only_config: GxConfig,
):
    """
    In order to load the seeded cloud config, this fixture must be called before any
    `get_context()` calls.
    """
    org_url_base = urllib.parse.urljoin(GX_CLOUD_MOCK_BASE_URL, f"organizations/{FAKE_ORG_ID}")

    fake_db_data = create_fake_db_seed_data(fds_config=fluent_only_config)
    _CLOUD_API_FAKE_DB.update(fake_db_data)

    seeded_datasources = _CLOUD_API_FAKE_DB["data-context-configuration"]["datasources"]
    CNF_TEST_LOGGER.info(f"Seeded Datasources ->\n{pf(seeded_datasources, depth=2)}")
    assert seeded_datasources

    yield cloud_api_fake

    assert len(cloud_api_fake.calls) >= 1, f"{org_url_base} was never called"


@pytest.fixture
def seeded_cloud_context(
    seed_cloud,  # NOTE: this fixture must be called before the CloudDataContext is created
    empty_cloud_context_fluent,
):
    empty_cloud_context_fluent._init_datasources()
    return empty_cloud_context_fluent


@pytest.fixture(
    params=[
        pytest.param("seeded_file_context", marks=[pytest.mark.filesystem]),
        pytest.param("seeded_cloud_context", marks=[pytest.mark.cloud]),
    ]
)
def seeded_contexts(
    request: FixtureRequest,
):
    """Parametrized fixture for seeded File and Cloud DataContexts."""
    context_fixture: FileDataContext | CloudDataContext = request.getfixturevalue(request.param)
    return context_fixture


@pytest.fixture
def pandas_filesystem_datasource(empty_data_context) -> PandasFilesystemDatasource:
    base_directory_rel_path = pathlib.Path("..", "..", "test_sets", "taxi_yellow_tripdata_samples")
    base_directory_abs_path = (
        pathlib.Path(__file__).parent.joinpath(base_directory_rel_path).resolve(strict=True)
    )
    pandas_filesystem_datasource = PandasFilesystemDatasource(
        name="pandas_filesystem_datasource",
        base_directory=base_directory_abs_path,
    )
    pandas_filesystem_datasource._data_context = empty_data_context
    return pandas_filesystem_datasource


@contextmanager
def _source(
    validate_batch_spec: Callable[[SqlAlchemyDatasourceBatchSpec], None],
    dialect: str,
    connection_string: str = "postgresql+psycopg2://postgres:@localhost/test_ci",
    data_context: Optional[AbstractDataContext] = None,
    partitioner_query_response: Optional[List[Dict[str, Any]]] = None,
    create_temp_table: bool = True,
) -> Generator[PostgresDatasource, None, None]:
    partitioner_response = partitioner_query_response or (
        [
            {"year": year, "month": month}
            for year in _DEFAULT_TEST_YEARS
            for month in _DEFAULT_TEST_MONTHS
        ]
    )

    execution_eng_cls = sqlachemy_execution_engine_mock_cls(
        validate_batch_spec=validate_batch_spec,
        dialect=dialect,
        partitioner_query_response=partitioner_response,
    )
    original_override = PostgresDatasource.execution_engine_override  # type: ignore[misc] # FIXME CoP
    try:
        PostgresDatasource.execution_engine_override = execution_eng_cls  # type: ignore[misc] # FIXME CoP
        postgres_datasource = PostgresDatasource(
            name="my_datasource",
            connection_string=connection_string,
            create_temp_table=create_temp_table,
        )
        if data_context:
            postgres_datasource._data_context = data_context
        yield postgres_datasource
    finally:
        PostgresDatasource.execution_engine_override = original_override  # type: ignore[misc] # FIXME CoP


# We may be able to parameterize this fixture so we can instantiate _source in the fixture.
# This would reduce the `with ...` boilerplate in the individual tests.
@pytest.fixture
def create_source() -> ContextManager:
    return _source  # type: ignore[return-value] # FIXME CoP


@pytest.fixture
def sql_datasource_table_asset_test_connection_noop(
    ephemeral_context_with_defaults: EphemeralDataContext,
    monkeypatch: pytest.MonkeyPatch,
    filter_gx_datasource_warnings: None,
) -> SQLDatasource:
    """
    SQLDatasource instance where `TableAsset.test_connection()` always passes.
    """
    sql_datasource = ephemeral_context_with_defaults.data_sources.add_sql(
        name="my_sql_datasource", connection_string="sqlite:///"
    )
    CNF_TEST_LOGGER.warning(f"Patching {sql_datasource.name} `.test_connection()` to a noop")

    def noop(self: SQLDatasource | TableAsset):
        CNF_TEST_LOGGER.warning(f"{self.__class__.__name__}.test_connection noop called")

    monkeypatch.setattr(TableAsset, "test_connection", noop, raising=True)
    return sql_datasource



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    Union,
)

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.id_dict import IDDict

if TYPE_CHECKING:
    from great_expectations.core.metric_function_types import (
        MetricFunctionTypes,
        MetricPartialFunctionTypes,
    )
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation import Expectation
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.expectations.metrics.metric_provider import MetricProvider
    from great_expectations.render import (
        AtomicDiagnosticRendererType,
        AtomicPrescriptiveRendererType,
        AtomicRendererType,
        RenderedAtomicContent,
        RenderedContent,
    )
    from great_expectations.validator.computed_metric import MetricValue

logger = logging.getLogger(__name__)

_registered_expectations: dict = {}
_registered_metrics: dict = {}
_registered_renderers: dict = {}

"""
{
  "metric_name"
    metric_domain_keys
    metric_value_keys
    metric_dependencies
    providers:
      engine: provider
}
"""


class RendererImpl(NamedTuple):
    expectation: str
    renderer: Callable[
        ..., Union[RenderedAtomicContent, list[RenderedAtomicContent], RenderedContent]
    ]


def register_renderer(
    object_name: str,
    parent_class: Union[Type[Expectation], Type[MetricProvider]],
    renderer_fn: Callable[
        ..., Union[RenderedAtomicContent, list[RenderedAtomicContent], RenderedContent]
    ],
):
    # noinspection PyUnresolvedReferences
    renderer_name = renderer_fn._renderer_type  # type: ignore[attr-defined] # FIXME CoP
    if object_name not in _registered_renderers:
        logger.debug(f"Registering {renderer_name} for expectation_type {object_name}.")
        _registered_renderers[object_name] = {renderer_name: (parent_class, renderer_fn)}
        return

    if renderer_name in _registered_renderers[object_name]:
        if _registered_renderers[object_name][renderer_name] == (
            parent_class,
            renderer_fn,
        ):
            logger.info(
                f"Multiple declarations of {renderer_name} renderer for expectation_type {object_name} "  # noqa: E501 # FIXME CoP
                f"found."
            )
            return
        else:
            logger.warning(
                f"Overwriting declaration of {renderer_name} renderer for expectation_type "
                f"{object_name}."
            )
            _registered_renderers[object_name][renderer_name] = (
                parent_class,
                renderer_fn,
            )
        return
    else:
        logger.debug(f"Registering {renderer_name} for expectation_type {object_name}.")
        _registered_renderers[object_name][renderer_name] = (parent_class, renderer_fn)
        return


def get_renderer_names(expectation_or_metric_type: str) -> List[str]:
    """Gets renderer names for a given Expectation or Metric.

    Args:
        expectation_or_metric_type: The type of an Expectation or Metric for which to get renderer names.

    Returns:
        A list of renderer names for the Expectation or Metric.
    """  # noqa: E501 # FIXME CoP
    return list(_registered_renderers.get(expectation_or_metric_type, {}).keys())


def get_renderer_names_with_renderer_types(
    expectation_or_metric_type: str,
    renderer_types: List[AtomicRendererType],
) -> List[Union[str, AtomicDiagnosticRendererType, AtomicPrescriptiveRendererType]]:
    """Gets renderer names of a given type, for a given Expectation or Metric.

    Args:
        expectation_or_metric_type: The type of an Expectation or Metric for which to get renderer names.
        renderer_types: The type of the renderers for which to return names.

    Returns:
        A list of renderer names for the given prefixes and Expectation or Metric.
    """  # noqa: E501 # FIXME CoP
    return [
        renderer_name
        for renderer_name in get_renderer_names(
            expectation_or_metric_type=expectation_or_metric_type
        )
        if any(renderer_name.startswith(renderer_type) for renderer_type in renderer_types)
    ]


def get_renderer_impls(object_name: str) -> List[str]:
    return list(_registered_renderers.get(object_name, {}).values())


def get_renderer_impl(object_name: str, renderer_type: str) -> Optional[RendererImpl]:
    renderer_tuple: Optional[tuple] = _registered_renderers.get(object_name, {}).get(renderer_type)
    renderer_impl: Optional[RendererImpl] = None
    if renderer_tuple:
        renderer_impl = RendererImpl(expectation=renderer_tuple[0], renderer=renderer_tuple[1])
    return renderer_impl


def register_expectation(expectation: Type[Expectation]) -> None:
    expectation_type = expectation.expectation_type
    # TODO: add version to key
    if expectation_type in _registered_expectations:
        if _registered_expectations[expectation_type] == expectation:
            logger.info(f"Multiple declarations of expectation {expectation_type} found.")
            return
        else:
            logger.warning(f"Overwriting declaration of expectation {expectation_type}.")

    logger.debug(f"Registering expectation: {expectation_type}")
    _registered_expectations[expectation_type] = expectation


def register_core_metrics() -> None:
    """As Metric registration is the responsibility of MetaMetricProvider.__new__,
    simply importing a given class will ensure that it is added to the Metric
    registry.

    We use this to grab metrics by name within our workflows.

    Without this function, we need to hope that core Metrics are imported somewhere
    in our import graph - if not, our registry will be empty and we'll see
    MetricResolutionErrors.
    """
    before_count = len(_registered_metrics)

    # Implicitly calls MetaMetricProvider.__new__ as Metrics are loaded from metrics.__init__.py
    # As __new__ calls upon register_metric this import builds our core registry
    from great_expectations.expectations import metrics  # noqa: F401 # FIXME CoP

    after_count = len(_registered_metrics)

    if before_count == after_count:
        logger.debug("Already registered core metrics; no updates to registry")
    else:
        logger.debug(f"Registered {after_count - before_count} core metrics")


def register_core_expectations() -> None:
    """As Expectation registration is the responsibility of MetaExpectation.__new__,
    simply importing a given class will ensure that it is added to the Expectation
    registry.

    We use this JIT in the Validator to ensure that core Expectations are available
    for usage when called upon.

    Without this function, we need to hope that core Expectations are imported somewhere
    in our import graph - if not, our registry will be empty and Validator workflows
    will fail.
    """
    before_count = len(_registered_expectations)

    # Implicitly calls MetaExpectation.__new__ as Expectations are loaded from core.__init__.py
    # As __new__ calls upon register_expectation, this import builds our core registry
    from great_expectations.expectations import core  # noqa: F401 # FIXME CoP

    after_count = len(_registered_expectations)

    if before_count == after_count:
        logger.debug("Already registered core expectations; no updates to registry")
    else:
        logger.debug(f"Registered {after_count - before_count} core expectations")


def _add_response_key(res, key, value):
    if key in res:
        res[key].append(value)
    else:
        res[key] = [value]
    return res


def register_metric(  # noqa: PLR0913 # FIXME CoP
    metric_name: str,
    metric_domain_keys: Tuple[str, ...],
    metric_value_keys: Tuple[str, ...],
    execution_engine: Type[ExecutionEngine],
    metric_class: Type[MetricProvider],
    metric_provider: Optional[Callable],
    metric_fn_type: Optional[Union[MetricFunctionTypes, MetricPartialFunctionTypes]] = None,
) -> dict:
    """Register a Metric class for use as a callable metric within Expectations.

    Args:
        metric_name: A name identifying the metric. Metric Name must be globally unique in
            a great_expectations installation.
        metric_domain_keys: A tuple of the keys used to determine the domain of the metric.
        metric_value_keys: A tuple of the keys used to determine the value of the metric.
        execution_engine: The execution_engine used to execute the metric.
        metric_class: A valid Metric class containing logic to compute attributes of data.
        metric_provider: The MetricProvider class from which the metric_class inherits.
        metric_fn_type: The MetricFunctionType or MetricPartialFunctionType used to define the Metric class.

    Returns:
        A dictionary containing warnings thrown during registration if applicable, and the success status of registration.
    """  # noqa: E501 # FIXME CoP
    res: dict = {}
    execution_engine_name = execution_engine.__name__
    logger.debug(f"Registering metric: {metric_name}")
    if metric_provider is not None and metric_fn_type is not None:
        metric_provider.metric_fn_type = metric_fn_type  # type: ignore[attr-defined] # FIXME CoP
    if metric_name in _registered_metrics:
        metric_definition = _registered_metrics[metric_name]
        current_domain_keys = metric_definition.get("metric_domain_keys", set())
        if set(current_domain_keys) != set(metric_domain_keys):
            logger.warning(
                f"metric {metric_name} is being registered with different metric_domain_keys; overwriting metric_domain_keys"  # noqa: E501 # FIXME CoP
            )
            _add_response_key(
                res,
                "warning",
                f"metric {metric_name} is being registered with different metric_domain_keys; overwriting metric_domain_keys",  # noqa: E501 # FIXME CoP
            )

        current_value_keys = metric_definition.get("metric_value_keys", set())
        if set(current_value_keys) != set(metric_value_keys):
            logger.warning(
                f"metric {metric_name} is being registered with different metric_value_keys; overwriting metric_value_keys"  # noqa: E501 # FIXME CoP
            )
            _add_response_key(
                res,
                "warning",
                f"metric {metric_name} is being registered with different metric_value_keys; overwriting metric_value_keys",  # noqa: E501 # FIXME CoP
            )

        providers = metric_definition.get("providers", {})
        if execution_engine_name in providers:
            _current_provider_cls, current_provider_fn = providers[execution_engine_name]
            if current_provider_fn != metric_provider:
                logger.warning(
                    f"metric {metric_name} is being registered with different metric_provider; overwriting metric_provider"  # noqa: E501 # FIXME CoP
                )
                _add_response_key(
                    res,
                    "warning",
                    f"metric {metric_name} is being registered with different metric_provider; overwriting metric_provider",  # noqa: E501 # FIXME CoP
                )
                providers[execution_engine_name] = metric_class, metric_provider
            else:
                logger.info(
                    f"Multiple declarations of metric {metric_name} for engine {execution_engine_name}."  # noqa: E501 # FIXME CoP
                )
                _add_response_key(
                    res,
                    "info",
                    f"Multiple declarations of metric {metric_name} for engine {execution_engine_name}.",  # noqa: E501 # FIXME CoP
                )
        else:
            providers[execution_engine_name] = metric_class, metric_provider
    else:
        metric_definition = {
            "metric_domain_keys": metric_domain_keys,
            "metric_value_keys": metric_value_keys,
            "default_kwarg_values": metric_class.default_kwarg_values,
            "providers": {execution_engine_name: (metric_class, metric_provider)},
        }
        _registered_metrics[metric_name] = metric_definition

    res["success"] = True

    return res


def _get_metric_definition(metric_name: str) -> dict:
    try:
        return _registered_metrics[metric_name]
    except KeyError:
        raise gx_exceptions.MetricProviderError(f"No metric named {metric_name} found.")  # noqa: TRY003 # FIXME CoP


def get_sqlalchemy_metric_provider(
    metric_name: str,
) -> Tuple[MetricProvider, Callable]:
    """The default SqlAlchemy metric provider for a given metric."""
    metric_definition = _get_metric_definition(metric_name)
    try:
        return metric_definition["providers"]["SqlAlchemyExecutionEngine"]
    except KeyError:
        raise gx_exceptions.MetricProviderError(  # noqa: TRY003 # FIXME CoP
            f"No provider found for {metric_name} using SqlAlchemyExecutionEngine"
        )


def get_metric_provider(
    metric_name: str, execution_engine: ExecutionEngine
) -> Tuple[MetricProvider, Callable]:
    metric_definition = _get_metric_definition(metric_name)
    try:
        return metric_definition["providers"][type(execution_engine).__name__]
    except KeyError:
        # Search up class hierarchy for a match. We skip the first entry since that's the
        # execution engine type itself, type(execution_engine), which we just checked and
        # resulted in the KeyError we're handling here.
        for cls in type(execution_engine).mro()[1:]:
            possible_key = cls.__name__
            if metric_definition["providers"].get(possible_key) is not None:
                metric_provider = metric_definition["providers"][possible_key]
                # Register the metric provider for this engine so we don't have to search again
                metric_definition["providers"][type(execution_engine).__name__] = metric_provider
                return metric_provider
        # no matches when search hierarchy so we raise
        raise gx_exceptions.MetricProviderError(  # noqa: TRY003 # FIXME CoP
            f"No provider found for {metric_name} using {type(execution_engine).__name__}"
        )


def get_metric_function_type(
    metric_name: str, execution_engine: ExecutionEngine
) -> Optional[Union[MetricPartialFunctionTypes, MetricFunctionTypes]]:
    try:
        provider_fn, _provider_class = get_metric_provider(metric_name, execution_engine)
        return getattr(provider_fn, "metric_fn_type", None)
    except KeyError:
        raise gx_exceptions.MetricProviderError(  # noqa: TRY003 # FIXME CoP
            f"No provider found for {metric_name} using {type(execution_engine).__name__}"
        )


def get_metric_kwargs(
    metric_name: str,
    configuration: Optional[ExpectationConfiguration] = None,
    runtime_configuration: Optional[dict] = None,
) -> dict:
    try:
        metric_definition = _registered_metrics.get(metric_name)
        if metric_definition is None:
            raise gx_exceptions.MetricProviderError(f"No definition found for {metric_name}")  # noqa: TRY003 # FIXME CoP
        default_kwarg_values = metric_definition["default_kwarg_values"]
        metric_kwargs = {
            "metric_domain_keys": metric_definition["metric_domain_keys"],
            "metric_value_keys": metric_definition["metric_value_keys"],
        }
        if configuration:
            expectation = configuration.to_domain_obj()
            configuration_kwargs = expectation._get_runtime_kwargs(
                runtime_configuration=runtime_configuration
            )
            if len(metric_kwargs["metric_domain_keys"]) > 0:
                metric_domain_kwargs = IDDict(
                    {
                        k: configuration_kwargs.get(k) or default_kwarg_values.get(k)
                        for k in metric_kwargs["metric_domain_keys"]
                    }
                )
            else:
                metric_domain_kwargs = IDDict()
            if len(metric_kwargs["metric_value_keys"]) > 0:
                metric_value_kwargs = IDDict(
                    {
                        k: configuration_kwargs.get(k)
                        if configuration_kwargs.get(k) is not None
                        else default_kwarg_values.get(k)
                        for k in metric_kwargs["metric_value_keys"]
                    }
                )
            else:
                metric_value_kwargs = IDDict()
            metric_kwargs["metric_domain_kwargs"] = metric_domain_kwargs
            metric_kwargs["metric_value_kwargs"] = metric_value_kwargs
        return metric_kwargs
    except KeyError:
        raise gx_exceptions.MetricProviderError(f"Incomplete definition found for {metric_name}")  # noqa: TRY003 # FIXME CoP


def get_domain_metrics_dict_by_name(
    metrics: Dict[Tuple[str, str, str], MetricValue], metric_domain_kwargs: IDDict
):
    return {
        metric_edge_key_id_tuple[0]: metric_value
        for metric_edge_key_id_tuple, metric_value in metrics.items()
        if metric_edge_key_id_tuple[1] == metric_domain_kwargs.to_id()
    }


def get_expectation_impl(expectation_name: str) -> Type[Expectation]:
    expectation: Type[Expectation] | None = _registered_expectations.get(expectation_name)
    if not expectation:
        raise gx_exceptions.ExpectationNotFoundError(f"{expectation_name} not found")  # noqa: TRY003 # FIXME CoP

    return expectation


def list_registered_expectation_implementations(
    expectation_root: Optional[Type[Expectation]] = None,
) -> List[str]:
    registered_expectation_implementations = []
    for (
        expectation_name,
        expectation_implementation,
    ) in _registered_expectations.items():
        if expectation_root is None or (
            expectation_root and issubclass(expectation_implementation, expectation_root)
        ):
            registered_expectation_implementations.append(expectation_name)

    return registered_expectation_implementations
